theory SCT
imports Ramsey Edg SubtermCriterionImpl UsableRules PolyOrder
begin

subsection {* Size change graphs and algorithms working on them *}

datatype ('pp, 'ap) scg =
 Null  -- {* zero element to make composition total *}
| Scg 'pp 'pp "('ap * 'ap) list" "('ap * 'ap) list"

text {* Composition *}

definition comp :: "('ap * 'ap) list \<Rightarrow> ('ap * 'ap) list \<Rightarrow> ('ap * 'ap) list"
where "comp es es' = remdups [ (x, z). (x,y) \<leftarrow> es, (y',z) \<leftarrow> es', y = y' ]"

lemma set_comp[simp]:
 "set (comp es es') = set es' O set es"
unfolding comp_def by (auto split: if_splits intro!: rel_compI) force

definition "subtract_list xs ys = filter (\<lambda>x. x \<notin> set ys) xs"
lemma set_subtract_list[simp]: 
  "set (subtract_list xs ys) = set xs - set ys"
  unfolding subtract_list_def by auto

text {*
 Scg composition depends on an (executable) relation @{term conn} that describes
 connectedness of program points
*}

fun scg_comp :: "('pp \<Rightarrow> 'pp \<Rightarrow> bool) \<Rightarrow> ('pp, 'ap) scg \<Rightarrow> ('pp, 'ap) scg \<Rightarrow> ('pp, 'ap) scg"
where
 "scg_comp conn (Scg p q str wk) (Scg p' q' str' wk') =
    (if \<not>conn q p' then Null else
     let strs = remdups (comp str str' @ comp str wk' @ comp wk str');
         wks = subtract_list (remdups (comp wk wk')) strs
     in Scg p q' strs wks)"
| "scg_comp conn Null G = Null"
| "scg_comp conn G Null = Null"

text {* Subsumption *}

fun subsumes :: "('pp, 'ap) scg \<Rightarrow> ('pp, 'ap) scg \<Rightarrow> bool"
where
 "subsumes (Scg p q str wk) (Scg p' q' str' wk') =
    (p = p' \<and> q = q' \<and> set str \<subseteq> set str' \<and> set wk \<subseteq> set (str' @ wk'))"
| "subsumes G Null = True"
| "subsumes Null G = False"

lemma subsumes_refl[simp]: "subsumes G G"
by (cases G) auto

lemma subsumes_trans[trans]:
 "subsumes F G \<Longrightarrow> subsumes G H \<Longrightarrow> subsumes F H"
by (cases H, simp, induct F G rule: subsumes.induct) auto

lemma subsumes_comp_mono:
 "subsumes G G' \<Longrightarrow> subsumes H H'
 \<Longrightarrow> subsumes (scg_comp conn G H) (scg_comp conn G' H')"
apply (induct G G' rule: subsumes.induct)
apply auto
apply (induct H H' rule: subsumes.induct)
apply (auto simp: Let_def)
apply blast+
done

definition subsume :: "('pp, 'ap) scg set \<Rightarrow> ('pp, 'ap) scg set \<Rightarrow> bool"
where
 "subsume Gs Hs = (\<forall>H\<in>Hs. H \<noteq> Null \<longrightarrow> (\<exists>G\<in>Gs. subsumes G H))"

lemma subsume_refl[simp]: "subsume Gs Gs"
by (force simp: subsume_def)

lemma subsume_trans[trans]:
  assumes "subsume Fs Gs" "subsume Gs Hs"
  shows "subsume Fs Hs"
unfolding subsume_def
proof (intro ballI impI)
  fix H assume "H \<in> Hs" "H \<noteq> Null"
  with `subsume Gs Hs`
  obtain G where G: "G \<in> Gs" "subsumes G H" unfolding subsume_def by auto
  with `H \<noteq> Null` have "G \<noteq> Null" by (cases G, cases H, auto)
  with `subsume Fs Gs` G
  obtain F where F: "F \<in> Fs" "subsumes F G" unfolding subsume_def by auto
  from F G show "\<exists>F\<in>Fs. subsumes F H" using subsumes_trans by auto
qed

lemma subsume_Un[simp]:
 "subsume Fs (Gs \<union> Hs) = (subsume Fs Gs \<and> subsume Fs Hs)"
by (auto simp: subsume_def)

lemma subsume_insert:
 "subsume Fs (insert G Gs) = (subsume Fs {G} \<and> subsume Fs Gs)"
by (auto simp: subsume_def)

text {* Transitivity check *}

definition trans_scgs
where
 "trans_scgs conn Gs =
   (\<forall>G\<in>set Gs. \<forall>H\<in>set Gs.
     let GH = scg_comp conn G H
     in (\<exists>G'\<in>set Gs. subsumes G' GH))"

subsubsection {* Checking the Scgs for descent *}

text {* Check for in-situ descent *}

primrec in_situ :: "('pp, 'ap) scg \<Rightarrow> bool"
where
 "in_situ Null = True"
| "in_situ (Scg p q str wk) = (\<exists>(x,y)\<in>set str. x = y)"

lemma subsumes_in_situ:
 "subsumes G H \<Longrightarrow> in_situ G \<Longrightarrow> in_situ H"
by (induct G H rule: subsumes.induct) auto

text {* Sagiv's test (aka cycles check): Some power of G has a strict descent: *}

primrec exp :: "('pp \<Rightarrow> 'pp \<Rightarrow> bool) \<Rightarrow> ('pp, 'ap) scg \<Rightarrow> nat \<Rightarrow> ('pp, 'ap) scg"
where
 "exp conn G 0 = G"
| "exp conn G (Suc n) = scg_comp conn (exp conn G n) G"

lemma exp_Null[simp]: "exp conn Null n = Null"
by (induct n) auto 


definition sagiv
where
 [code del]: "sagiv conn G = (\<exists>n. in_situ (exp conn G n))"

text {* Executable version: *}

fun combine :: "('pp, 'ap) scg \<Rightarrow> ('pp, 'ap) scg \<Rightarrow> ('pp, 'ap) scg"
where
 "combine (Scg p q str wk) (Scg p' q' str' wk') =
    Scg p q (remdups (str @ str')) (remdups (wk @ wk'))"
| "combine Null S = Null"
| "combine S Null = Null"

lemma in_situ_combine: "in_situ (combine G H) = (in_situ G \<or> in_situ H)"
by (induct G H rule: combine.induct) auto

inductive scg_eqv where
  "scg_eqv Null Null"
| "set str = set str' \<Longrightarrow> set str \<union> set wk = set str' \<union> set wk'
  \<Longrightarrow> scg_eqv (Scg p q str wk) (Scg p q str' wk')"

lemma equiv_classes_eq:
  fixes R :: "'a \<Rightarrow> 'a \<Rightarrow> bool" (infix "\<approx>" 70)
  assumes refl: "\<And>x. x \<approx> x"
  assumes trans: "\<And>x y z. x \<approx> y \<Longrightarrow> y \<approx> z \<Longrightarrow> x \<approx> z"
  assumes sym: "\<And>x y. x \<approx> y \<Longrightarrow> y \<approx> x"
  shows "x \<approx> y = (R x = R y)"
proof
  { fix x y assume "x \<approx> y"
    have "R x \<subseteq> R y"
    proof
      fix z assume "z \<in> R x" hence "x \<approx> z" unfolding mem_def .
      from trans[OF sym[OF this] `x \<approx> y`, THEN sym]
      show "z \<in> R y" unfolding mem_def .
    qed
  } note sub = this
  assume "x \<approx> y"
  show "R x = R y"
    by (intro equalityI sub[OF `x \<approx> y`] sub[OF sym[OF `x \<approx> y`]])
next
  assume "R x = R y" thus "x \<approx> y" by (simp add: refl)
qed

lemma scg_eqv: "scg_eqv G H = (scg_eqv G = scg_eqv H)"
proof (rule equiv_classes_eq)
  fix G show "scg_eqv G G"
    by (cases G) (auto intro: scg_eqv.intros)
next
  fix F G H assume "scg_eqv F G" "scg_eqv G H"
  thus "scg_eqv F H" by (auto elim!: scg_eqv.cases intro!: scg_eqv.intros)
next
  fix F G assume "scg_eqv F G"
  thus "scg_eqv G F" by (auto elim!: scg_eqv.cases intro!: scg_eqv.intros)
qed

lemma scg_eqv1: "scg_eqv G = scg_eqv H \<Longrightarrow> scg_eqv G H"
by (rule iffD2[OF scg_eqv])
lemma scg_eqv2: "scg_eqv G H \<Longrightarrow> scg_eqv G = scg_eqv H"
by (rule iffD1[OF scg_eqv])

lemma comp_assoc: 
  "scg_eqv (scg_comp conn (scg_comp conn F G) H)
  (scg_comp conn F (scg_comp conn G H))"
proof (cases "scg_comp conn (scg_comp conn F G) H")
  case Null
  thus ?thesis using scg_eqv.intros[intro]
    by (cases F, auto, cases G, auto, cases H, auto simp: Let_def)
next
  case Scg
  then obtain p1 q1 s1 w1 p2 q2 s2 w2 p3 q3 s3 w3
    where "F = Scg p1 q1 s1 w1"
    "G = Scg p2 q2 s2 w2"
    "H = Scg p3 q3 s3 w3" "conn q1 p2" "conn q2 p3"
    by (cases F, simp, cases G, simp, cases H, 
      simp_all add: Let_def split: if_splits, blast)

  thus ?thesis 
    apply (simp add: Let_def)
    apply (rule scg_eqv.intros)
    apply auto
    apply blast+
    done
qed

lemma comp_cong: 
assumes G: "scg_eqv G = scg_eqv G'" and H: "scg_eqv H = scg_eqv H'"
shows "scg_eqv (scg_comp conn G H) = scg_eqv (scg_comp conn G' H')"
proof (cases G)
  case Null 
  from G[THEN scg_eqv1] have Null': "G' = Null"
    by (rule scg_eqv.cases) (auto simp: Null)
  show ?thesis 
    by (rule scg_eqv2) (auto simp: Null Null' intro: scg_eqv.intros)
next  
  case (Scg p q s w)
  note scg' = this
  with G[THEN scg_eqv1] obtain sa wa
    where G': "G' = Scg p q sa wa"
    and G_eq: "set s = set sa" "set s \<union> set w = set sa \<union> set wa"
    by (auto elim!: scg_eqv.cases)

  show ?thesis
  proof (cases H)
    case Null
    from H[THEN scg_eqv1] have Null': "H' = Null"
      by (rule scg_eqv.cases) (auto simp: Null)
    show ?thesis 
      by (rule scg_eqv2) (auto simp: scg' G' Null Null' intro: scg_eqv.intros)
  next
    case (Scg p' q' s' w')
    with H[THEN scg_eqv1] obtain sa' wa'
      where H': "H' = Scg p' q' sa' wa'"
      and H_eq: "set s' = set sa'"
      "set s' \<union> set w' = set sa' \<union> set wa'"
      by (auto elim!: scg_eqv.cases)

    show ?thesis
    proof (cases "conn q p'")
      case True
 
      have "(set s' \<union> set w') O set s
        \<union> set s' O (set s \<union> set w)
        = (set sa' \<union> set wa') O set sa
        \<union> set sa' O (set sa \<union> set wa)"
        by (simp only: G_eq(2) H_eq(2), simp only: G_eq H_eq)
      then have 1:
        "(set s' O set s \<union> (set w' O set s \<union> set s' O set w))
=        (set sa' O set sa \<union> (set wa' O set sa \<union> set sa' O set wa))"
  by (simp add: Un_ac)

      have 2: "(set s' \<union> set w') O (set s \<union> set w) 
        = (set sa' \<union> set wa') O (set sa \<union> set wa)"
        by (simp only: G_eq(2) H_eq(2))

      show ?thesis unfolding scg' Scg G' H'
        apply (rule scg_eqv2)
        apply (simp add: Let_def True)
        apply (rule scg_eqv.intros)
        apply (simp add: set_remdups set_append set_comp 1)
        apply (insert 2, simp add: Un_ac)
        done
      next
      case False
      show ?thesis unfolding scg' Scg G' H'
        by (rule scg_eqv2)
      (auto simp: Let_def False intro: scg_eqv.intros)
    qed
  qed
qed


lemma exp_cong: "scg_eqv G = scg_eqv H 
  \<Longrightarrow> scg_eqv (exp conn G n) = scg_eqv (exp conn H n)"
by (induct n) (auto intro!: comp_cong)

lemma exp_comm: "scg_eqv (scg_comp conn G (exp conn G n))
  = scg_eqv (scg_comp conn (exp conn G n) G)"
proof (induct n)
  case 0 show ?case by simp
next
  case (Suc n)
  show ?case
    apply simp
    apply (unfold comp_assoc[THEN scg_eqv2, symmetric])
    apply (rule comp_cong)
    apply (unfold Suc[symmetric])
    by auto
qed

lemma exp_compose:
  "scg_eqv (scg_comp conn (exp conn G n) (exp conn G m))
  = scg_eqv (exp conn G (Suc (n + m)))"
proof (induct n)
  case 0 show ?case by (simp add: exp_comm)
next
  case (Suc n)
  have "scg_eqv (scg_comp conn (scg_comp conn (exp conn G n) G) (exp conn G m)) 
    = scg_eqv (scg_comp conn (exp conn G n) (scg_comp conn G (exp conn G m)))"
    by (simp add: comp_assoc[THEN scg_eqv2])
  also have "... = scg_eqv (scg_comp conn (exp conn G n) (scg_comp conn (exp conn G m) G))"
    by (rule comp_cong) (auto simp: exp_comm)
  also have "... = scg_eqv (scg_comp conn (scg_comp conn (exp conn G n) (exp conn G m)) G)"
    by (simp add: comp_assoc[THEN scg_eqv2])
  also have  "... = scg_eqv (scg_comp conn (exp conn G (Suc (n + m))) G)"   
    by (rule comp_cong) (auto simp: Suc)
  finally show ?case by simp
qed

primrec strict_edge :: "'ap \<Rightarrow> 'ap \<Rightarrow> ('pp, 'ap) scg \<Rightarrow> bool"
where
  "strict_edge i j Null = True"
| "strict_edge i j (Scg _ _ str _) = ((i,j)\<in>set str)"

primrec weak_edge :: "'ap \<Rightarrow> 'ap \<Rightarrow> ('pp, 'ap) scg \<Rightarrow> bool"
where
  "weak_edge i j Null = True"
| "weak_edge i j (Scg _ _ str wk) = ((i,j)\<in>set wk \<union> set str)"

lemma edge_combine[simp]: 
  "strict_edge i j (combine G H) = (strict_edge i j G \<or> strict_edge i j H)"
  "weak_edge i j (combine G H) = (weak_edge i j G \<or> weak_edge i j H)"
by (induct G H rule: combine.induct) auto

lemma weak_edge_decompose: 
  assumes "scg_comp conn G H \<noteq> Null"
  assumes "weak_edge i k (scg_comp conn G H)"
  obtains j where "weak_edge i j G" "weak_edge j k H"
using assms
by (induct conn G H rule: scg_comp.induct) (auto simp: Let_def split: if_splits)

lemma strict_edge_decompose: 
  assumes "scg_comp conn G H \<noteq> Null"
  assumes "strict_edge i k (scg_comp conn G H)"
  obtains j where "strict_edge i j G" "strict_edge j k H"
  | j where "strict_edge i j G" "weak_edge j k H"
  | j where "weak_edge i j G" "strict_edge j k H"
using assms
by (induct conn G H rule: scg_comp.induct) (auto simp: Let_def split: if_splits)


lemma exp_compose_strict: "strict_edge i j (scg_comp conn (exp conn G n) (exp conn G m))
 = strict_edge i j (exp conn G (Suc (n + m)))"
using exp_compose[THEN scg_eqv1, of conn G n m]
by rule auto 

lemma exp_compose_weak: "weak_edge i j (scg_comp conn (exp conn G n) (exp conn G m))
 = weak_edge i j (exp conn G (Suc (n + m)))"
using exp_compose[THEN scg_eqv1, of conn G n m]
by rule auto 

lemma subsumes_exp_mono:
assumes "subsumes G H"
  shows "subsumes (exp conn G n) (exp conn H n)"
by (induct n) (auto intro: subsumes_comp_mono assms)

lemma combine_subsumes: "subsumes G (combine G H)"
by (induct G H rule: combine.induct) auto

lemma in_situ_subsumes:
  "subsumes G H \<Longrightarrow> in_situ G \<Longrightarrow> in_situ H"
by (induct G H rule: subsumes.induct) auto

lemma sagiv_subsumes: 
assumes sub: "subsumes G H" and G: "sagiv conn G"
shows "sagiv conn H"
proof -
  from G obtain n where 1: "in_situ (exp conn G n)" unfolding sagiv_def ..
  from sub have "subsumes (exp conn G n) (exp conn H n)" 
    by (rule subsumes_exp_mono)
  from this 1 have "in_situ (exp conn H n)" by (rule in_situ_subsumes)
  thus ?thesis unfolding sagiv_def ..
qed

lemma sagiv_step:
  "sagiv conn G \<longleftrightarrow> sagiv conn (combine G (scg_comp conn G G))"
  (is "?A \<longleftrightarrow> ?B")
proof
  assume ?A 
  then show ?B by (rule sagiv_subsumes[OF combine_subsumes])
next
  let ?G' = "combine G (scg_comp conn G G)"
  assume ?B show ?A
  proof (cases ?G')
    case Null 
    hence "G = Null \<or> scg_comp conn G G = Null" by (cases G) (auto simp: Let_def)
    thus ?thesis
    proof
      assume "G = Null" thus ?thesis 
        by (auto simp: sagiv_def)
    next
      assume "scg_comp conn G G = Null"
      hence "in_situ (exp conn G 1)" by auto
      thus ?thesis unfolding sagiv_def ..
    qed
  next
    case (Scg p q str wk)
    then obtain str' wk' where Scg': "G = Scg p q str' wk'" 
      by (cases G) (auto split: if_splits simp: Let_def)
    with Scg have composable: "conn q p" by (auto split: if_splits)
    {
      fix n
      from composable Scg
      have "\<exists>s w. exp conn ?G' n = Scg p q s w" by (induct n) (auto simp: Let_def)
      hence "exp conn ?G' n \<noteq> Null" by auto
    }
    note nonnull = this

    {
      fix i j
      assume "strict_edge i j ?G'"
      from this[simplified]
      have "\<exists>m. strict_edge i j (exp conn G m)"
      proof
        assume "strict_edge i j G" then have "strict_edge i j (exp conn G 0)" by auto
        thus ?thesis ..
      next
        assume "strict_edge i j (scg_comp conn G G)" then have "strict_edge i j (exp conn G 1)" by auto
        thus ?thesis ..
      qed
    }
    note strict_step = this

    {
      fix i j
      assume "weak_edge i j ?G'"
      from this[simplified]
      have "\<exists>m. weak_edge i j (exp conn G m)"
      proof
        assume "weak_edge i j G" then have "weak_edge i j (exp conn G 0)" by auto
        thus ?thesis ..
      next
        assume "weak_edge i j (scg_comp conn G G)" then have "weak_edge i j (exp conn G 1)" by auto
        thus ?thesis ..
      qed
    }
    note weak_step = this


    {
      fix i j n
      have "strict_edge i j (exp conn ?G' n) \<Longrightarrow> \<exists>m. strict_edge i j (exp conn G m)"
       and "weak_edge i j (exp conn ?G' n) \<Longrightarrow> \<exists>m. weak_edge i j (exp conn G m)"
      proof (induct n arbitrary: i j)
        case 0 case 1 thus ?case by (simp add: strict_step)
      next
        case 0 case 2
         hence "weak_edge i j G \<or> weak_edge i j (scg_comp conn G G)" 
          (is "?X \<or> ?Y") by auto
        thus ?case
        proof
          assume ?X then have "weak_edge i j (exp conn G 0)" by auto
          thus ?thesis ..
        next
          assume ?Y then have "weak_edge i j (exp conn G 1)" by auto
          thus ?thesis ..
        qed
      next
        case (Suc n) case (1 i k)
        then have a: "strict_edge i k (scg_comp conn (exp conn ?G' n) ?G')" by simp
        show ?case
        proof (rule strict_edge_decompose[OF nonnull[of "Suc n", simplified] a])
          fix j
          assume 1: "strict_edge i j (exp conn ?G' n)"
             and 2: "strict_edge j k ?G'"
          from Suc(1)[OF 1] strict_step[OF 2]
          obtain m m' 
            where "strict_edge i j (exp conn G m)" "strict_edge j k (exp conn G m')"
            by auto
          then have "strict_edge i k (scg_comp conn (exp conn G m) (exp conn G m'))"
            by (cases "exp conn G m", auto, cases "exp conn G m'", auto simp: Let_def)
          then have "strict_edge i k (exp conn G (Suc (m + m')))" unfolding exp_compose_strict .
          thus "\<exists>m. strict_edge i k (exp conn G m)" ..
        next
          fix j
          assume 1: "strict_edge i j (exp conn ?G' n)"
             and 2: "weak_edge j k ?G'"
          from Suc(1)[OF 1] weak_step[OF 2]
          obtain m m' 
            where "strict_edge i j (exp conn G m)" "weak_edge j k (exp conn G m')"
            by auto
          then have "strict_edge i k (scg_comp conn (exp conn G m) (exp conn G m'))"
            by (cases "exp conn G m", auto, cases "exp conn G m'", auto simp: Let_def)
          then have "strict_edge i k (exp conn G (Suc (m + m')))" unfolding exp_compose_strict .
          thus "\<exists>m. strict_edge i k (exp conn G m)" ..
        next
          fix j
          assume 1: "weak_edge i j (exp conn ?G' n)"
             and 2: "strict_edge j k ?G'"
          from Suc(2)[OF 1] strict_step[OF 2]
          obtain m m' 
            where "weak_edge i j (exp conn G m)" "strict_edge j k (exp conn G m')"
            by auto
          then have "strict_edge i k (scg_comp conn (exp conn G m) (exp conn G m'))"
            apply (cases "exp conn G m", auto)
            apply (cases "exp conn G m'", auto simp: Let_def)
            apply (cases "exp conn G m'", auto simp: Let_def) done
          then have "strict_edge i k (exp conn G (Suc (m + m')))" unfolding exp_compose_strict .
          thus "\<exists>m. strict_edge i k (exp conn G m)" ..
        qed        
      next
        case (Suc n) case (2 i k)
        then have a: "weak_edge i k (scg_comp conn (exp conn ?G' n) ?G')" by simp
        show ?case
        proof (rule weak_edge_decompose[OF nonnull[of "Suc n", simplified] a])
          fix j
          assume 1: "weak_edge i j (exp conn ?G' n)"
             and 2: "weak_edge j k ?G'"
          from Suc(2)[OF 1] weak_step[OF 2]
          obtain m m' 
            where "weak_edge i j (exp conn G m)" "weak_edge j k (exp conn G m')"
            by auto
          then have "weak_edge i k (scg_comp conn (exp conn G m) (exp conn G m'))"
            apply (cases "exp conn G m", auto)
            apply (cases "exp conn G m'", auto simp: Let_def)
            apply (cases "exp conn G m'", auto simp: Let_def) done
          then have "weak_edge i k (exp conn G (Suc (m + m')))" unfolding exp_compose_weak .
          thus "\<exists>m. weak_edge i k (exp conn G m)" ..
        qed        
      qed       
    }
    note r = this(1)
    from `?B` obtain n
      where "in_situ (exp conn ?G' n)"
      unfolding sagiv_def .. 
    with nonnull
    obtain i 
      where "strict_edge i i (exp conn ?G' n)"
      by (cases "exp conn (combine G (scg_comp conn G G)) n") auto
    from r[OF this]
    obtain m where "strict_edge i i (exp conn G m)" ..
    then
    have "in_situ (exp conn G m)"
      by (cases "exp conn G m") auto
    thus "sagiv conn G"
      unfolding sagiv_def ..
  qed
qed

lemma sagiv_code[code]:
 "sagiv conn G =
  (if in_situ G
     then True
     else let GG = scg_comp conn G G
          in if subsumes GG G then False
             else sagiv conn (combine G GG))"
proof cases
 assume G: "in_situ G"
 hence "in_situ (exp conn G 0)" by simp
 hence "sagiv conn G" unfolding sagiv_def ..
 with G show ?thesis by simp
next
 assume nG: "\<not> in_situ G"
 show ?thesis
 proof cases
   assume sub: "subsumes (scg_comp conn G G) G"
   {
     fix n have "subsumes (exp conn G n) G"
     proof (induct n)
       case 0 show ?case by simp
     next
       case (Suc n)
       from this subsumes_refl
       have "subsumes (scg_comp conn (exp conn G n) G) (scg_comp conn G G)"
         by (rule subsumes_comp_mono)
       also have "subsumes (\<dots>) G" by (fact sub)
       finally show ?case by simp
     qed
   }
   from subsumes_in_situ[OF this] and nG
   have "\<not> sagiv conn G" unfolding sagiv_def by auto
   with nG sub show ?thesis by auto
 next
   assume nsub: "\<not> subsumes (scg_comp conn G G) G"
   with nG sagiv_step
   show ?thesis by auto
 qed
qed


text {* Transitive closure computation *}

definition insert_scg :: "('pp, 'ap) scg \<Rightarrow> ('pp, 'ap) scg list \<Rightarrow> ('pp, 'ap) scg list option"
where
 "insert_scg G Gs =
  (if G = Null then None
   else if (\<exists>G'\<in>set Gs. subsumes G' G) then None
   else Some (G # [ G' \<leftarrow> Gs. \<not> subsumes G G' ]))"

lemma insert_scg_None:
 "insert_scg G Gs = None \<Longrightarrow> subsume (set Gs) {G}"
by (auto simp: insert_scg_def subsume_def split: if_splits)

lemma insert_scg_Some1:
 "insert_scg G Gs = Some Gs' \<Longrightarrow> subsume (set Gs') ({G} \<union> set Gs)"
by (force simp: insert_scg_def subsume_def split: if_splits)

lemma insert_scg_Some2:
 "insert_scg G Gs = Some Gs' \<Longrightarrow> subsume ({G} \<union> set Gs) (set Gs') \<and> G \<in> set Gs'"
by (force simp: insert_scg_def subsume_def split: if_splits)

function mk_trancl :: "('pp \<Rightarrow> 'pp \<Rightarrow> bool) \<Rightarrow> ('pp, 'ap) scg list \<Rightarrow> ('pp, 'ap) scg list \<Rightarrow> ('pp, 'ap) scg list"
where
 "mk_trancl conn [] S = S"
| "mk_trancl conn (G#Gs) S =
  (case insert_scg G S of
     None \<Rightarrow> mk_trancl conn Gs S
   | Some S' \<Rightarrow> mk_trancl conn (map (scg_comp conn G) S' @ map (\<lambda>F. scg_comp conn F G) S' @ Gs) S')"
by pat_completeness auto


text {* Termination proof *}

fun pps_of :: "('pp, 'ap) scg \<Rightarrow> 'pp set"
where
  "pps_of Null = {}"
| "pps_of (Scg p q _ _) = {p,q}"
fun aps_of :: "('pp, 'ap) scg \<Rightarrow> 'ap set"
where
  "aps_of Null = {}"
| "aps_of (Scg _ _ str wk) = set (map fst str) \<union> set (map snd str) \<union> set (map fst wk) \<union> set (map snd wk)"

lemma finite_pps[simp]: "finite (pps_of G)" by (cases G) auto
lemma finite_aps[simp]: "finite (aps_of G)" by (cases G) auto

lemma image_fst_snd: "(a,b)\<in>S \<Longrightarrow> a \<in> fst ` S" "(a,b)\<in>S \<Longrightarrow> b \<in> snd ` S"
by force+

lemma pps_of_comp: "pps_of (scg_comp conn G H) \<subseteq> pps_of G \<union> pps_of H"
by (induct conn G H rule: scg_comp.induct) (auto simp: Let_def)
lemma aps_of_comp: "aps_of (scg_comp conn G H) \<subseteq> aps_of G \<union> aps_of H"
by (induct conn G H rule: scg_comp.induct) (auto simp: image_fst_snd Let_def)

definition graph_bound where
  "graph_bound T =
  (let P = (\<Union>H\<in>T. pps_of H); A = (\<Union>H\<in>T. aps_of H)
   in P \<times> P \<times> Pow (A \<times> A) \<times> Pow (A \<times> A))"

lemma graph_bound_subsetI:
"(\<Union>G\<in>S. pps_of G) \<subseteq> (\<Union>G\<in>T. pps_of G)
\<Longrightarrow> (\<Union>G\<in>S. aps_of G) \<subseteq> (\<Union>G\<in>T. aps_of G)
\<Longrightarrow> graph_bound S \<subseteq> graph_bound T"
unfolding graph_bound_def by auto

lemma graph_bound_union2:
  assumes "graph_bound A \<subseteq> graph_bound C"
  assumes "graph_bound B \<subseteq> graph_bound C"
  shows "graph_bound (A \<union> B) \<subseteq> graph_bound C"
proof (rule graph_bound_subsetI)
  from assms have A: "graph_bound A \<union> graph_bound B \<subseteq> graph_bound C" by auto
  show "(\<Union>G\<in>A \<union> B. pps_of G) \<subseteq> (\<Union>G\<in>C. pps_of G)" (is "?P1 \<subseteq> ?P2")
  proof
    fix p assume "p \<in> ?P1"
    then obtain G where AB: "G \<in> A \<union> B" and p: "p \<in> pps_of G" by auto
    then have "(p,p,{},{}) \<in> graph_bound A \<union> graph_bound B"
      unfolding graph_bound_def by auto
    with A have "(p,p,{},{}) \<in> graph_bound C" by auto
    thus "p \<in> ?P2" unfolding graph_bound_def by auto
  qed
  show "(\<Union>G\<in>A \<union> B. aps_of G) \<subseteq> (\<Union>G\<in>C. aps_of G)" (is "?A1 \<subseteq> ?A2")
  proof
    fix a assume "a \<in> ?A1"
    then obtain G where AB: "G \<in> A \<union> B" and a: "a \<in> aps_of G" by auto
    then obtain p where p: "p \<in> pps_of G" by (cases G) auto
    with AB a have "(p,p,{(a,a)},{(a,a)}) \<in> graph_bound A \<union> graph_bound B"
      unfolding graph_bound_def by auto
    with A have "(p,p,{(a,a)},{(a,a)}) \<in> graph_bound C" by auto
    thus "a \<in> ?A2" unfolding graph_bound_def by auto
  qed
qed

lemma graph_bound_mono: "Gs \<subseteq> Hs \<Longrightarrow> graph_bound Gs \<subseteq> graph_bound Hs"
by (rule graph_bound_subsetI) auto

lemma graph_boundI:
  assumes "G = Scg p q str wk"
  shows "(p, q, set str, set wk) \<in> graph_bound {G}"
  unfolding graph_bound_def assms assms Let_def
  by (auto simp: image_fst_snd)

lemma pps_insert:
  "insert_scg G S = Some S' \<Longrightarrow> (\<Union>H\<in>set S'. pps_of H) \<subseteq> (\<Union>H\<in>insert G (set S). pps_of H)"
by (auto simp: insert_scg_def split: if_splits)

lemma aps_insert:
  "insert_scg G S = Some S' \<Longrightarrow> (\<Union>H\<in>set S'. aps_of H) \<subseteq> (\<Union>H\<in>insert G (set S). aps_of H)"
by (auto simp: insert_scg_def split: if_splits)

lemma graph_bound_insert:
  "insert_scg G S = Some S'
  \<Longrightarrow> graph_bound (set S') \<subseteq> graph_bound (insert G (set S))"
by (intro graph_bound_subsetI conjI pps_insert aps_insert)

fun abs_graph where
  "abs_graph Null = {}"
| "abs_graph (Scg p q str wk) = {(p',q',S',W').
  p=p' \<and> q=q' \<and> set str \<subseteq> S' \<and> set wk \<subseteq> S' \<union> W'}"

definition abs_graphs where
  "abs_graphs S = (\<Union>G\<in>set S. abs_graph G)"

lemma abs_graph_subsumes: "subsumes G H \<Longrightarrow> abs_graph H \<subseteq> abs_graph G"
by (induct G H rule: subsumes.induct) auto

lemma abs_graphs_subsume:
  assumes sub: "subsume (set Gs) (set Hs)"
  shows "abs_graphs Hs \<subseteq> abs_graphs Gs"
  unfolding abs_graphs_def
proof (rule UN_least)
  fix H assume "H \<in> set Hs"
  with sub have "H = Null \<or> (\<exists>G\<in>set Gs. subsumes G H)"
    by (auto simp: subsume_def)
  with abs_graph_subsumes
  show "abs_graph H \<subseteq> UNION (set Gs) abs_graph"
    by force
qed


inductive_set saturation for f bound R
where
  strict: "f A \<inter> bound A \<subset> f B \<inter> bound A \<Longrightarrow> finite (bound A) \<Longrightarrow> bound B \<subseteq> bound A
  \<Longrightarrow> (B, A) \<in> saturation f bound R"
| nonstrict: "f A \<inter> bound A \<subseteq> f B \<inter> bound A \<Longrightarrow> finite (bound A) \<Longrightarrow> bound B \<subseteq> bound A
\<Longrightarrow> (B, A) \<in> R \<Longrightarrow> (B, A) \<in> saturation f bound R"

lemma wf_saturation: "wf R \<Longrightarrow> wf (saturation f bound R)"
apply (rule wf_subset[where r="(\<lambda>x. card (bound x - f x)) <*mlex*> R"])
apply (rule wf_mlex, assumption)
apply auto
apply (erule saturation.cases)
apply (rule mlex_less)
apply simp
apply (rule psubset_card_mono)
apply simp
apply blast
apply (rule mlex_leq)
apply (rule card_mono)
apply simp
apply blast
apply simp
done

termination mk_trancl
proof
  let ?R = "saturation (\<lambda>(_,W,S). abs_graphs S)
                           (\<lambda>(_,W,S). graph_bound (set W \<union> set S))
              (measure (\<lambda>(_,W,S). length W))"

  show "wf ?R" by (auto simp: wf_saturation)

  fix conn :: "'a \<Rightarrow> 'a \<Rightarrow> bool" and G and Gs S :: "('a, 'b) scg list"

  show "((conn, Gs, S), (conn, G#Gs, S)) \<in> ?R"
    by (rule saturation.nonstrict) (auto simp add: graph_bound_def)

  fix S'
  assume ins: "insert_scg G S = Some S'"

  from insert_scg_Some1[OF ins]
  have "subsume (set S') (set S)" unfolding subsume_def by auto
  then have 2: "abs_graphs S \<subseteq> abs_graphs S'" by (rule abs_graphs_subsume)
  from ins have 3: "\<not>(\<exists>H \<in> set S. subsumes H G)" unfolding insert_scg_def
    by (auto split: if_splits)

  from ins obtain p q str wk
    where Scg: "G = Scg p q str wk" by (cases G, auto simp: insert_scg_def)

  let ?aG = "(p, q, set str, set wk)"
  let ?B = "insert G (set Gs \<union> set S)"
  have "?aG \<notin> abs_graphs S"
  proof
    assume "?aG \<in> abs_graphs S"
    then obtain H where "H \<in> set S" "?aG \<in> abs_graph H"
      unfolding abs_graphs_def by auto
    with Scg have "subsumes H G" by (cases H) auto
    with 3 `H \<in> set S` show False by auto
  qed
  moreover
  have "?aG \<in> graph_bound ?B"
    using graph_boundI[OF Scg] graph_bound_mono[of "{G}"] by auto
  moreover
  from insert_scg_Some2[OF ins]
  have "G \<in> set S'" ..
  with Scg
  have "?aG \<in> abs_graphs S'" unfolding abs_graphs_def by force
  ultimately
  have 3: "abs_graphs S \<inter> graph_bound ?B \<subset> abs_graphs S' \<inter> graph_bound ?B" using 2 by auto

  from graph_bound_insert[OF ins]
  have "graph_bound (set S')
    \<subseteq> graph_bound (insert G (set S))" .
  also have "\<dots> \<subseteq> graph_bound (insert G (set Gs \<union> set S))"
    by (rule graph_bound_mono) auto
  finally have A: "graph_bound (set S')
    \<subseteq> graph_bound (insert G (set Gs \<union> set S))" .

  have "graph_bound (scg_comp conn G ` set S' \<union> (\<lambda>F. scg_comp conn F G) ` set S')
    \<subseteq> graph_bound (insert G (set S'))"
    apply (rule graph_bound_union2)
    apply (rule graph_bound_subsetI)
    defer defer
    apply (rule graph_bound_subsetI)
    by (auto dest: subsetD[OF pps_of_comp] subsetD[OF aps_of_comp])
    also have "\<dots> \<subseteq> graph_bound (insert G (set Gs \<union> set S))"
      apply (subst insert_is_Un)
      apply (rule graph_bound_union2)
      apply (rule graph_bound_mono, simp)
      by (rule A)
    finally have B: "graph_bound (scg_comp conn G ` set S' \<union> (\<lambda>F. scg_comp conn F G) ` set S')
    \<subseteq> graph_bound (insert G (set Gs \<union> set S))" .

  have bound_smaller: "graph_bound
     (scg_comp conn G ` set S' \<union> ((\<lambda>F. scg_comp conn F G) ` set S' \<union>  set Gs) \<union> set S')
    \<subseteq> graph_bound (insert G (set Gs \<union> set S))"
    apply (rule graph_bound_union2)
    apply (simp only: Un_assoc[symmetric])
    apply (rule graph_bound_union2)
    apply (rule B)
    apply (rule graph_bound_mono, blast)
    apply (rule A)
    done

  show "((conn, map (scg_comp conn G) S' @
                  map (\<lambda>F. scg_comp conn F G) S' @ Gs, S'),
         (conn, G # Gs, S)) \<in> ?R"
    apply (rule saturation.strict)
    apply simp_all
    apply (rule 3)
    apply (simp add: graph_bound_def)
    apply (rule bound_smaller)
    done
qed

text {* correctness of @{term "mk_trancl"} *}

lemma trancl_closure:
 "subsume (set (mk_trancl conn W S)) (set W \<union> set S)"
proof (induct conn W S rule: mk_trancl.induct)
 case 1 thus ?case by simp
next
 case (2 _ G Gs S)
 show ?case
 proof (cases "insert_scg G S")
   case None
   with insert_scg_None[OF None] 2
   show ?thesis
     by (auto simp: subsume_insert[of _ _ "set Gs \<union> set S"] dest: subsume_trans)
 next
   case (Some Sa)
   with insert_scg_Some1[OF Some] 2(2)[OF Some]
   show ?thesis
     by (auto simp: subsume_insert[of _ _ "set S"] subsume_insert[of _ _ "set Gs \<union> set S"]
         dest: subsume_trans)
 qed
qed

lemma trancl_trans:
 assumes "subsume (set S \<union> set W) ({ scg_comp conn F G |F G. F \<in> set S \<and> G \<in> set S })"
         (is "?inv conn S W")
 shows "trans_scgs conn (mk_trancl conn W S)"
 using assms
proof (induct conn W S rule: mk_trancl.induct)
 case (1 _ S) thus ?case unfolding trans_scgs_def Let_def
   by (force simp: subsume_def)
next
 case (2 conn G Gs S)
 hence i: "?inv conn S (G#Gs)" by simp
 show ?case
 proof (cases "insert_scg G S")
   case None
   from insert_scg_None[OF None]
   have "subsume (set S \<union> set Gs) (set S \<union> set (G # Gs))"
     by (force simp: subsume_def)
   hence "?inv conn S Gs" using i by (rule subsume_trans)
   from 2(1)[OF None this] None
   show ?thesis by auto
 next
   case (Some Sa)
   have SS: "\<And>F Ga. F \<in> set S \<Longrightarrow> Ga \<in> set S \<Longrightarrow>
          \<exists>H. subsumes H (scg_comp conn F Ga) \<and> (H \<in> set S \<or> H \<in> set Gs \<or> H = G)"
   proof -
     fix F Ga assume "F \<in> set S" "Ga \<in> set S"
     then have ss: "scg_comp conn F Ga \<in> {scg_comp conn F G |F G. F \<in> set S \<and> G \<in> set S}"
       (is "?FGa \<in> ?SS") by auto

     show "\<exists>H. subsumes H (scg_comp conn F Ga)
       \<and> (H \<in> set S \<or> H \<in> set Gs \<or> H = G)"
     proof (cases "?FGa")
       case Null thus ?thesis by auto
     next
       case Scg
       hence "?FGa \<noteq> Null" by simp
       with i ss show ?thesis
         by (auto simp: subsume_def)
     qed
   qed

   from  insert_scg_Some2[OF Some]
   have Sab: "\<And>F. F \<in> set Sa \<Longrightarrow> subsumes G F \<or> (\<exists>H\<in>set S. subsumes H F)"
     and G: "G \<in> set Sa"
     by (auto simp: subsume_def)

   have "?inv conn Sa (map (scg_comp conn G) Sa @ map (\<lambda>F. scg_comp conn F G) Sa @ Gs)"
   proof (auto simp: subsume_def)
     fix F Ga assume "F \<in> set Sa" "Ga \<in> set Sa"
     with Sab have "subsumes G F \<or> (\<exists>H\<in>set S. subsumes H F)" by simp
     thus "\<exists>G\<in>set Sa \<union>
                (scg_comp conn G ` set Sa \<union>
                 ((\<lambda>F. scg_comp conn F G) ` set Sa \<union> set Gs)).
                subsumes G (scg_comp conn F Ga)"
     proof
       assume "subsumes G F"
       from this subsumes_refl have "subsumes (scg_comp conn G Ga) (scg_comp conn F Ga)" 
         by (rule subsumes_comp_mono)
       moreover from `Ga \<in> set Sa` have "scg_comp conn G Ga \<in> scg_comp conn G ` set Sa" by auto
       ultimately show ?thesis by auto
     next
       assume "\<exists>H\<in>set S. subsumes H F"
       then obtain H where H: "H \<in> set S" and "subsumes H F" by auto
       from `F \<in> set Sa` obtain H' where H': "H' \<in> set Sa" and "subsumes H' F" by auto
       from Sab[OF `Ga \<in> set Sa`]
       show ?thesis
       proof
         assume "subsumes G Ga"
         with subsumes_refl have "subsumes (scg_comp conn F G) (scg_comp conn F Ga)" 
           by (rule subsumes_comp_mono)
         moreover from `F \<in> set Sa` have "scg_comp conn F G \<in> (\<lambda>F. scg_comp conn F G) ` set Sa" by auto
         ultimately show ?thesis by auto
       next
         assume "\<exists>H\<in>set S. subsumes H Ga"
         then obtain H2 where H2: "H2 \<in> set S" and "subsumes H2 Ga" by auto
         with `subsumes H F` have "subsumes (scg_comp conn H H2) (scg_comp conn F Ga)" by (auto intro: subsumes_comp_mono)
         with SS[OF H H2] obtain H3 where sH3: "subsumes H3 (scg_comp conn F Ga)" 
           and H3: "H3 \<in> set S \<or> H3 \<in> set Gs \<or> H3 = G" by (auto dest: subsumes_trans)
         from H3 have "H3 \<in> {G} \<union> set S \<or> H3 \<in> set Gs" by auto
         thus ?thesis
         proof
           assume "H3 \<in> {G} \<union> set S"
           with sH3 have "subsume ({G} \<union> set S) {scg_comp conn F Ga}" by (auto simp: subsume_def)
           with insert_scg_Some1[OF Some]
           have "subsume (set Sa) {scg_comp conn F Ga}" by (rule subsume_trans)
           with `F \<in> set Sa`
           obtain H4 where "H4 \<in> set Sa" "subsumes H4 (scg_comp conn F Ga)"
             by (auto simp: subsume_def)
           thus ?thesis by auto
         next
           assume "H3 \<in> set Gs"
           with sH3 show ?thesis by auto
         qed
       qed
     qed
   qed
   from 2(2)[OF Some this] Some
   show ?thesis by auto
 qed
qed

text {* All-in-one SCT check *}

definition check_SCT :: "('pp \<Rightarrow> 'pp \<Rightarrow> bool) \<Rightarrow> ('pp, 'ap) scg list \<Rightarrow> bool"
where
 "check_SCT conn Gs =
    (let S = mk_trancl conn Gs []
     in (\<forall>G\<in>set S. sagiv conn G))"


subsubsection {* Code generation *}

lemma [code unfold]: "(set xs \<subseteq> set ys) = (\<forall>x\<in>set xs. x\<in>set ys)"
by (fact subset_eq)


subsection {* Transition system semantics *}

locale sct_semantics =
 fixes S NS :: "'a brel"
 fixes conn :: "'pp \<Rightarrow> 'pp \<Rightarrow> bool"
 assumes wf_S: "SN S"
 assumes SN_compat: "S O NS \<subseteq> S"
begin

abbreviation "trN == (NS \<union> S)^*"
abbreviation "trS == trN O S O trN"

lemma tr_subset: "trS \<subseteq> trN"
proof -
 have "trS \<subseteq> trN O trN O trN"
   by (intro rel_comp_mono) auto
 also have "\<dots> \<subseteq> trN" by auto
 finally show ?thesis .
qed

lemma SN_compat': "trS O trN \<subseteq> trS"
 by (simp add: O_assoc)
lemma NS_compat': "trN O trS \<subseteq> trS"
 by (simp add: O_assoc[symmetric])
lemma SS_compat': "trS O trS \<subseteq> trS"
 using tr_subset NS_compat' by blast
lemma NN_compat': "trN O trN \<subseteq> trN"
 by auto

lemma SN_trS: "SN trS"
using SN_compat wf_S by (rule rtrancl_S_rtrancl_is_SN)

primrec steps :: "('pp, 'ap) scg \<Rightarrow> ('pp \<times> ('ap \<Rightarrow> 'a)) brel"
where
 "steps Null = {}"
| "steps (Scg p q str wk) =
  { ((p, f) , (r, g)) | f g r.
      (\<forall>(x,y)\<in>set str. (f x, g y) \<in> trS)
    \<and> (\<forall>(x,y)\<in>set wk. (f x, g y) \<in> trN) \<and> conn q r }"


lemma steps_subsumes: "subsumes G G' \<Longrightarrow> steps G' \<subseteq> steps G"
using tr_subset
by (induct G G' rule: subsumes.induct) (auto, blast+)

lemma steps_comp:
shows "steps G' O steps G \<subseteq> steps (scg_comp conn G G')"
proof (induct "c::'pp\<Rightarrow>'pp\<Rightarrow>bool" G G' rule: scg_comp.induct)
 case (1 _ p q str wk p' q' str' wk')
 thus ?case
 proof (rule, simp_all only: split_paired_all)
   fix pa qa f g
   assume "((pa,f), (qa,g))
         \<in> steps (Scg p' q' str' wk') O steps (Scg p q str wk)"
   then obtain r h where
   "((pa,f), (r,h)) \<in> steps (Scg p q str wk)"
   "((r,h), (qa,g)) \<in> steps (Scg p' q' str' wk')"
     by auto
   thus "((pa,f), (qa,g))
         \<in> steps (scg_comp conn (Scg p q str wk) (Scg p' q' str' wk'))"
     apply (auto simp: subtract_list_def)
     apply (rule subsetD[OF SS_compat'])
     apply (rule_tac b="h y" in rel_compI, force, force)
     apply (rule subsetD[OF NS_compat'])
     apply (rule_tac b="h y" in rel_compI, force, force)
     apply (rule subsetD[OF SN_compat'])
     apply (rule_tac b="h y" in rel_compI, force, force)
     apply (rule subsetD[OF NN_compat'])
     apply (rule_tac b="h y" in rel_compI, force, force)
     done
 qed
qed auto

lemma trans_scgs:
 assumes tr: "trans_scgs conn Gs"
 shows "trans (\<Union>G\<in>set Gs. steps G)" (is "trans ?R")
proof (rule transI, simp only: split_paired_all)
 fix p q r f g h
 assume "((p, f), (q, g)) \<in> ?R" "((q, g), (r, h)) \<in> ?R"
 then obtain G H where GH: "G \<in> set Gs" "H \<in> set Gs"
   and "((p, f), (q, g)) \<in> steps G"
   "((q, g), (r, h)) \<in> steps H"
   by auto
 hence comp: "((p, f), (r, h)) \<in> steps (scg_comp conn G H)" using steps_comp by auto
 from tr GH obtain G' where "G' \<in> set Gs"
   and "subsumes G' (scg_comp conn G H)"
   unfolding trans_scgs_def by auto
 with comp have "((p, f), (r, h)) \<in> steps G'" using steps_subsumes by auto
 with `G' \<in> set Gs` show "((p, f), (r, h)) \<in> ?R" by auto
qed

lemma in_situ_SN:
 "in_situ G \<Longrightarrow> SN (steps G)"
proof (induct G)
 case (Scg p q str wk)
 then obtain x where "(x,x) \<in> set str" by auto
 then have subset: "steps (Scg p q str wk) \<subseteq> inv_image trS (\<lambda>(p,f). f x)"
   unfolding inv_image_def by auto
 have "SN (inv_image trS (\<lambda>(p,f). f x))"
   by (intro SN_inv_image SN_trS)
 then show ?case using subset
   by (rule SN_subset)
qed auto

lemma steps_exp: "steps G ^ Suc n \<subseteq> steps (exp conn G n)"
proof (induct n)
 case (Suc n)
 with steps_comp[of G "exp conn G n"]
 show ?case by auto
qed auto

lemma sagiv_SN: 
  assumes a: "sagiv conn G" 
  shows "SN (steps G)"
using a
unfolding sagiv_def
proof 
 fix n assume "in_situ (exp conn G n)"
 hence "SN (steps (exp conn G n))" by (rule in_situ_SN)
 from this steps_exp
 have "SN (steps G ^ Suc n)" by (rule SN_subset)
 thus "SN (steps G)" unfolding SN_pow[symmetric] .
qed

text {* Version that assumes that the closure is already given *}

lemma SCT_correctness:
 fixes R :: "('pp \<times> ('ap \<Rightarrow> 'a)) brel"
 assumes abst:  "R \<subseteq> (\<Union>G\<in>set Gs. steps G)"
 assumes trans: "trans_scgs conn Ts"
 assumes super: "subsume (set Ts) (set Gs)"
 assumes good:  "\<forall>G\<in>set Ts. sagiv conn G"
 shows "SN R"
proof -
 from good
 have "\<And>G. G \<in> set Ts \<Longrightarrow> SN (steps G)"
   by (auto dest!: sagiv_SN)
 hence disj: "disj_wf ((\<Union>G\<in>set Ts. steps G)^-1)"
   unfolding disj_wf_def
   apply (rule_tac x = "\<lambda>i. (steps (Ts ! i))^-1" in exI)
   apply (rule_tac x = "length Ts" in exI)
   apply (simp add: SN_iff_wf)
   apply auto
   apply (metis in_set_conv_nth lessThan_iff mem_def)
   apply (metis in_set_conv_nth)
   done
 from trans
 have "trans (\<Union>G\<in>set Ts. steps G)"
   by (rule trans_scgs)
 from this and disj
 have SN: "SN (\<Union>G\<in>set Ts. steps G)"
   unfolding SN_iff_wf
   by (auto intro: trans_disj_wf_implies_wf)
 note abst
 also from super have
   "(\<Union>G\<in>set Gs. steps G) \<subseteq> (\<Union>G\<in>set Ts. steps G)"
   by (force simp: subsume_def dest!: steps_subsumes)
 finally
 show "SN R" by (rule SN_subset[OF SN])
qed

text {* Version that computes the closure *}

lemma SCT_correctness2:
 fixes R :: "('pp \<times> ('ap \<Rightarrow> 'a)) brel"
 assumes abst:  "R \<subseteq> (\<Union>G\<in>set Gs. steps G)"
 assumes check: "check_SCT conn Gs"
 shows "SN R"
proof -
 let ?S = "mk_trancl conn Gs []"
 note abst
 moreover
 have "trans_scgs conn ?S" by (rule trancl_trans) (auto simp: subsume_def)
 moreover from trancl_closure[of conn Gs "[]"]
 have "subsume (set ?S) (set Gs)" by simp
 moreover from check[unfolded check_SCT_def Let_def]
 have "\<forall>G\<in>set ?S. sagiv conn G" by simp
 ultimately show "SN R"
   by (rule SCT_correctness)
qed
end

context redpair
begin

lemma sct_semantics: "sct_semantics S NS"
  by (rule sct_semantics.intro, rule SN, rule compat)

lemma sct_with_redpair:
   fixes P :: "('f,'v)trs" and R :: "('f,'v)trsL" and Gs :: "(('f,'v)rule, nat) scg list"
   assumes chain: "ichain (P,set R) s t \<sigma>"
   and no_left_vars: "\<forall> lr \<in> set R. \<forall> x. fst lr \<noteq> Var x"
   and tuple: "\<forall> (s,t) \<in> P. \<not> is_var s \<and> \<not> is_var t \<and> the (root t) \<notin> defs (set R)"
   and rules: "set R \<subseteq> NS"
   and graphs: "\<forall> (s,t) \<in> P. \<exists> stri non_stri. (
         Scg (s,t) (s,t) stri non_stri \<in> set Gs \<and>
         (\<forall> (i,j) \<in> set stri. i < length (args s) \<and> j < length (args t) \<and> (args s ! i, args t ! j) \<in> S) \<and> 
         (\<forall> (i,j) \<in> set non_stri. i < length (args s) \<and> j < length (args t) \<and> (args s ! i, args t ! j) \<in> NS))"
   and check: "check_SCT (\<lambda> st uv. checkEdgRM (computeRuleMap R) (computeRuleMap (reverse_trs R)) Q flag (st,uv)) Gs"
   shows False
proof -
  let ?pairs = "\<lambda> k. ((s k,t k), (\<lambda> j. args (s k) ! j \<cdot> \<sigma> k))"
  let ?Rs = "(rstep (set R))^*"
  let ?R = "{(?pairs k, ?pairs (Suc k)) | k. True}"
  let ?conn = "\<lambda> st uv. checkEdgRM (computeRuleMap R) (computeRuleMap (reverse_trs R)) Q flag (st,uv)"
  let ?S = "(NS \<union> S)\<^sup>* O S O (NS \<union> S)\<^sup>*"
  let ?NS = "(NS \<union> S)\<^sup>*"
  have main: "?R \<subseteq> (\<Union>G \<in> set Gs. sct_semantics.steps S NS ?conn G)"
  proof (simp, clarify, simp)
    fix k
    let ?sk = "\<lambda> i. args (s k) ! i \<cdot> \<sigma> k"
    let ?tk = "\<lambda> i. args (t k) ! i \<cdot> \<sigma> k"
    let ?sk1 = "\<lambda> i. args (s (Suc k)) ! i \<cdot> \<sigma> (Suc k)"
    from chain have inP: "(s k, t k) \<in> P" by auto
    with tuple have validDP: "\<not> is_var (t k) \<and> the (root (t k)) \<notin> defs (set R)" by auto
    from chain inP tuple have validDP1: "\<not> is_var (s (Suc k))" by auto
    from chain have steps: "(t k \<cdot> \<sigma> k, s (Suc k) \<cdot> \<sigma> (Suc k)) \<in> ?Rs" by auto
    have argSteps: "\<And> j. j < length (args (t k)) \<longrightarrow> (?tk j, ?sk1 j) \<in> ?Rs"
    proof
      fix j
      let ?tj = "args (t k) ! j"
      let ?sj = "args (s (Suc k)) ! j"
      from validDP obtain f ts where tk: "t k = Fun f ts \<and> f \<notin> defs (set R)" by (cases "t k", auto)
      hence f: "f \<notin> defs (set R)" ..
      from validDP1 obtain g ss where sk1: "s (Suc k) = Fun g ss" by (cases "s (Suc k)", auto) 
      assume j: "j < length (args (t k))"
      let ?tss = "(map (\<lambda> t. t \<cdot> \<sigma> k) ts)"
      let ?sss = "(map (\<lambda> s. s \<cdot> \<sigma> (Suc k)) ss)"
      from steps have "(Fun f ts \<cdot> \<sigma> k, Fun g ss \<cdot> \<sigma> (Suc k)) \<in> ?Rs" using tk sk1 by auto
      hence steps2: "(Fun f ?tss, Fun g ?sss) \<in> ?Rs" by auto
      from this no_left_vars f have "\<exists> us. length us = length ?tss \<and> Fun g ?sss = Fun f us \<and> (\<forall> i < length ?tss. (?tss ! i, us ! i) \<in> ?Rs)"
        by (rule nondef_root_imp_arg_steps)
      from this obtain us where nearly: "length us = length ?tss \<and> Fun g ?sss = Fun f us \<and> (\<forall> i < length ?tss. (?tss ! i, us ! i) \<in> ?Rs)" ..
      from nearly j tk sk1 show "(?tj \<cdot> \<sigma> k, ?sj \<cdot> \<sigma> (Suc k)) \<in> ?Rs" by auto
    qed
    from subNS wmNS rules trs_inclusion have "rstep (set R) \<subseteq> NS \<union> S" by blast
    hence stepsNS: "(rstep (set R))^* \<subseteq> ?NS" by (rule rtrancl_mono)
    from steps have conn: "?conn (s k, t k) (s (Suc k), t (Suc k))" by simp
    from inP obtain stri non_stri where graph: "Scg (s k, t k) (s k,t k) stri non_stri \<in> set Gs \<and> 
         (\<forall> (i,j) \<in> set stri. i < length (args (s k)) \<and> j < length (args (t k)) \<and> (args (s k) ! i, args (t k) ! j) \<in> S) \<and> 
         (\<forall> (i,j) \<in> set non_stri. i < length (args (s k)) \<and> j < length (args (t k)) \<and> (args (s k) ! i, args (t k) ! j) \<in> NS)" using graphs by force
    let ?g = "Scg (s k, t k) (s k, t k) stri non_stri"
    from graph have gGs: "?g \<in> set Gs" ..
    let ?check = "\<lambda> g. (((s k, t k), ?sk), (s (Suc k),t (Suc k)), ?sk1)
           \<in> sct_semantics.steps S NS ?conn g"
    have "?check ?g"
    proof (simp add: sct_semantics.steps.simps sct_semantics conn, rule conjI)
      {
        fix i j
        assume "(i,j) \<in> set non_stri"
        with graph have i: "i < length (args (s k))" and j: "j < length (args (t k))" and NS: "(args (s k) ! i, args (t k) ! j) \<in> NS" by auto
        from j argSteps have "(?tk j, ?sk1 j) \<in> ?Rs" by auto
        with stepsNS have astepsNS: "(?tk j, ?sk1 j) \<in> ?NS" by auto
        from NS subNS have non_strict: "(?sk i, ?tk j) \<in> (NS \<union> S)" unfolding sub_closed_def using csI by blast
        have "(?sk i, ?sk1 j) \<in> ?NS" by (rule rtrancl_trans, rule r_into_rtrancl, rule non_strict, rule astepsNS)
      }
      thus "\<forall>ij\<in>set non_stri.(\<lambda>(i, j). (?sk i, ?sk1 j) \<in> ?NS) ij" by auto
    next
      {
        fix i j
        assume "(i,j) \<in> set stri"
        with graph have i: "i < length (args (s k))" and j: "j < length (args (t k))" and S: "(args (s k) ! i, args (t k) ! j) \<in> S" by auto
        from j argSteps have "(?tk j, ?sk1 j) \<in> ?Rs" by auto
        with stepsNS have astepsNS: "(?tk j, ?sk1 j) \<in> ?NS" by auto
        from S subS have strict: "(?sk i, ?tk j) \<in> S" unfolding sub_closed_def using csI by blast
        have "(?sk i, ?sk1 j) \<in> ?S" using strict astepsNS by auto
      }
      thus "\<forall>ij\<in>set stri.(\<lambda>(i, j). (?sk i, ?sk1 j) \<in> ?S) ij" by auto
    qed
    with gGs show "\<exists> g \<in> set Gs. ?check g" by auto
  qed
  have "SN ?R"
    by (rule sct_semantics.SCT_correctness2[where S = S and NS = NS], rule sct_semantics, rule main, rule check)
  have "\<not> SN ?R" unfolding SN_ars_def SN_elt_def by force
  with `SN ?R` show False by auto
qed
end

context ce_af_redpair
begin
lemma sct_ur_af_redpair:
   fixes P :: "('a,'b)trs" and R :: "('a,'b)trsL" and Gs :: "(('a,'b)rule, nat) scg list"
   assumes chain: "min_ichain (P,set R) s t \<sigma>"
   and no_left_vars: "\<forall> lr \<in> set R. \<forall> x. fst lr \<noteq> Var x"
   and tuple: "\<forall> (s,t) \<in> P. \<not> is_var s \<and> \<not> is_var t \<and> the (root t) \<notin> defs (set R)"
   and rules: "ur \<subseteq> NS"
   and ur_closed: "urClosed_af ur R \<pi>"
   and graphs: "\<forall> (s,t) \<in> P. \<exists> stri non_stri. (
         Scg (s,t) (s,t) stri non_stri \<in> set Gs \<and>
         (\<forall> (i,j) \<in> set stri. i < length (args s) \<and> j < length (args t) \<and> urClosedT_af (args t ! j) R ur \<pi> \<and> (args s ! i, args t ! j) \<in> S) \<and> 
         (\<forall> (i,j) \<in> set non_stri. i < length (args s) \<and> j < length (args t) \<and> urClosedT_af (args t ! j) R ur \<pi> \<and> (args s ! i, args t ! j) \<in> NS))"
   and check: "check_SCT (\<lambda> st uv. checkEdgRM (computeRuleMap R) (computeRuleMap (reverse_trs R)) Q flag (st,uv)) Gs"
   shows False
proof -
  let ?I = "iTrans (R, ur, (c,d,n))"
  let ?Is = "\<lambda> i. (iTransSig (R,ur,(c,d,n)) (\<sigma> i))"
  let ?pairs = "\<lambda> k. ((s k,t k), (\<lambda> j. args (s k) ! j \<cdot> ?Is k))"
  let ?Rs = "(rstep (set R))^*"
  let ?R = "{(?pairs k, ?pairs (Suc k)) | k. True}"
  let ?conn = "\<lambda> st uv. checkEdgRM (computeRuleMap R) (computeRuleMap (reverse_trs R)) Q flag (st,uv)"
  let ?S = "(NS \<union> S)\<^sup>* O S O (NS \<union> S)\<^sup>*"
  let ?NS = "(NS \<union> S)\<^sup>*"
  have main: "?R \<subseteq> (\<Union>G \<in> set Gs. sct_semantics.steps S NS ?conn G)"
  proof (simp, clarify, simp)
    fix k
    let ?sk = "\<lambda> i. args (s k) ! i \<cdot> \<sigma> k"
    let ?tk = "\<lambda> i. args (t k) ! i \<cdot> \<sigma> k"
    let ?sk1 = "\<lambda> i. args (s (Suc k)) ! i \<cdot> \<sigma> (Suc k)"
    let ?Isk = "\<lambda> i. args (s k) ! i \<cdot> ?Is k"
    let ?Itk = "\<lambda> i. args (t k) ! i \<cdot> ?Is k"
    let ?Isk1 = "\<lambda> i. args (s (Suc k)) ! i \<cdot> ?Is (Suc k)"
    from chain have inP: "(s k, t k) \<in> P" by auto
    with tuple have validDP: "\<not> is_var (t k) \<and> the (root (t k)) \<notin> defs (set R)" by auto
    from chain inP tuple have validDP1: "\<not> is_var (s (Suc k))" by auto
    from chain have steps: "(t k \<cdot> \<sigma> k, s (Suc k) \<cdot> \<sigma> (Suc k)) \<in> ?Rs" by auto
    from chain have tSN: "SN (rstep (set R)) (t k \<cdot> \<sigma> k)" by (simp add: minimal_cond_def)
    have argSteps: "\<And> j. j < length (args (t k)) \<and> urClosedT_af (args (t k) ! j) R ur \<pi> \<longrightarrow> (?Itk j, ?Isk1 j) \<in> ?NS"
    proof
      fix j
      let ?tj = "args (t k) ! j"
      let ?sj = "args (s (Suc k)) ! j"
      from validDP obtain f ts where tk: "t k = Fun f ts \<and> f \<notin> defs (set R)" by (cases "t k", auto)
      hence f: "f \<notin> defs (set R)" ..
      from validDP1 obtain g ss where sk1: "s (Suc k) = Fun g ss" by (cases "s (Suc k)", auto) 
      assume "j < length (args (t k)) \<and> urClosedT_af ?tj R ur \<pi>" (is "?jlen \<and> ?urc")
      hence ?jlen and ?urc by auto
      let ?tss = "(map (\<lambda> t. t \<cdot> \<sigma> k) ts)"
      let ?sss = "(map (\<lambda> s. s \<cdot> \<sigma> (Suc k)) ss)"
      from tSN have SN: "SN (rstep (set R)) (Fun f ?tss)" using tk by auto
      from tk `?jlen` have mem: "?tj \<cdot> \<sigma> k \<in> set ?tss" by auto
      from steps have "(Fun f ts \<cdot> \<sigma> k, Fun g ss \<cdot> \<sigma> (Suc k)) \<in> ?Rs" using tk sk1 by auto
      hence steps2: "(Fun f ?tss, Fun g ?sss) \<in> ?Rs" by auto
      from this no_left_vars f have "\<exists> us. length us = length ?tss \<and> Fun g ?sss = Fun f us \<and> (\<forall> i < length ?tss. (?tss ! i, us ! i) \<in> ?Rs)"
        by (rule nondef_root_imp_arg_steps)
      from this obtain us where nearly: "length us = length ?tss \<and> Fun g ?sss = Fun f us \<and> (\<forall> i < length ?tss. (?tss ! i, us ! i) \<in> ?Rs)" ..
      from nearly `?jlen` tk sk1 have "(?tj \<cdot> \<sigma> k, ?sj \<cdot> \<sigma> (Suc k)) \<in> ?Rs" (is ?part1) by auto
      have SN2: "SN (rstep (set R)) (?tk j)"
        by  (rule SN_imp_SN_arg[where f = f and ts = ?tss], rule SN, rule mem)
      from SN2 `?part1` ur_closed rules have II: "(?I (?tk j), ?I (?sk1 j)) \<in> NS^*" by (rule iTransVI_af)
      from `?urc` SN2 have tI: "(?tj \<cdot> ?Is k, ?I (?tk j)) \<in> NS^*" by (rule iTransI_af)
      have Is: "(?I (?sk1 j), ?sj \<cdot> ?Is (Suc k)) \<in> NS^*" by (rule iTransII_af, rule SN_imp_SN_reduct, rule SN2, rule `?part1`)
      have "(?tj \<cdot> ?Is k, ?sj \<cdot> ?Is (Suc k)) \<in> NS^*" by (rule rtrancl_trans, rule tI, rule rtrancl_trans, rule II, rule Is)
      thus "(?tj \<cdot> ?Is k, ?sj \<cdot> ?Is (Suc k)) \<in> ?NS" using rtrancl_mono by blast
    qed
    from subNS wmNS rules trs_inclusion have urNS: "rstep ur \<subseteq> NS" by blast
    from steps have conn: "?conn (s k, t k) (s (Suc k), t (Suc k))" by simp
    from inP obtain stri non_stri where graph: "Scg (s k, t k) (s k,t k) stri non_stri \<in> set Gs \<and> 
         (\<forall> (i,j) \<in> set stri. i < length (args (s k)) \<and> j < length (args (t k)) \<and> urClosedT_af (args (t k) ! j) R ur \<pi> \<and> (args (s k) ! i, args (t k) ! j) \<in> S) \<and> 
         (\<forall> (i,j) \<in> set non_stri. i < length (args (s k)) \<and> j < length (args (t k)) \<and> urClosedT_af (args (t k) ! j) R ur \<pi> \<and> (args (s k) ! i, args (t k) ! j) \<in> NS)" using graphs by force
    let ?g = "Scg (s k, t k) (s k, t k) stri non_stri"
    from graph have gGs: "?g \<in> set Gs" ..
    let ?check = "\<lambda> g. (((s k, t k), ?Isk), (s (Suc k),t (Suc k)), ?Isk1)
           \<in> sct_semantics.steps S NS ?conn g"
    have "?check ?g"
    proof (simp add: sct_semantics.steps.simps sct_semantics conn, rule conjI)
      {
        fix i j
        assume "(i,j) \<in> set non_stri"
        with graph have i: "i < length (args (s k))" and j: "j < length (args (t k))"
          and usable: "urClosedT_af (args (t k) ! j) R ur \<pi>" and NS: "(args (s k) ! i, args (t k) ! j) \<in> NS" by auto
        from j usable argSteps have astepsNS: "(?Itk j, ?Isk1 j) \<in> ?NS" by auto
        from NS subNS have non_strict: "(?Isk i, ?Itk j) \<in> (NS \<union> S)" unfolding sub_closed_def using csI by blast
        have "(?Isk i, ?Isk1 j) \<in> ?NS" by (rule rtrancl_trans, rule r_into_rtrancl, rule non_strict, rule astepsNS) 
      }
      thus "\<forall>ij\<in>set non_stri.(\<lambda>(i, j). (?Isk i, ?Isk1 j) \<in> ?NS) ij" by auto
    next
      {
        fix i j
        assume "(i,j) \<in> set stri"
        with graph have i: "i < length (args (s k))" and j: "j < length (args (t k))"
          and usable: "urClosedT_af (args (t k) ! j) R ur \<pi>" and S: "(args (s k) ! i, args (t k) ! j) \<in> S" by auto
        from j usable argSteps have astepsNS: "(?Itk j, ?Isk1 j) \<in> ?NS" by auto
        from S subS have strict: "(?Isk i, ?Itk j) \<in> S" unfolding sub_closed_def using csI by blast
        have "(?Isk i, ?Isk1 j) \<in> ?S" using strict astepsNS by auto
      }
      thus "\<forall>ij\<in>set stri.(\<lambda>(i, j). (?Isk i, ?Isk1 j) \<in> ?S) ij" by auto
    qed
    with gGs show "\<exists> g \<in> set Gs. ?check g" by auto
  qed
  have "SN ?R"
    by (rule sct_semantics.SCT_correctness2[where S = S and NS = NS], rule sct_semantics, rule main, rule check)
  have "\<not> SN ?R" unfolding SN_ars_def SN_elt_def by force
  with `SN ?R` show False by auto
qed
end

lemma trancl_union_right[simp]: fixes r::"'a brel" shows "r^+ \<subseteq> (s \<union> r)^+"
proof (rule subsetI2)
  fix x y assume "(x,y) \<in> r^+" thus "(x,y) \<in> (s \<union> r)^+"
  proof (induct)
    case base thus ?case by auto
  next
    case (step a b)
    hence "(a,b) \<in> (s \<union> r)^+" by auto
    with `(x,a) \<in> (s \<union> r)^+` show ?case by auto
  qed
qed

lemma snRel_subset: "snRel R S \<subseteq> R"
proof (rule subsetI2)
  fix a b assume "(a,b) \<in> snRel R S" thus "(a,b) \<in> R" unfolding snRel_def by simp
qed

lemma supteq_supt_eq_conv: "supteq = supt \<union> Id" (is "?S = ?T") by auto

lemma sct_with_subterm:
   fixes P::"('f,'v)trs" and R::"('f,'v)trsL" and Gs::"(('f,'v)rule, nat) scg list"
   assumes chain: "min_ichain (P,set R) s t \<sigma>"
   and no_left_vars: "\<forall>lr\<in>set R.\<forall>x. fst lr \<noteq> Var x"
   and tuple: "\<forall>(s,t)\<in>P. \<not> is_var s \<and> \<not> is_var t \<and> the (root t) \<notin> defs (set R)"
   and graphs: "\<forall>(s,t)\<in>P.\<exists>stri nstri. (
         Scg (s,t) (s,t) stri nstri \<in> set Gs \<and>
         (\<forall>(i,j)\<in>set stri. i < length(args s) \<and> j < length(args t) \<and> (args s!i, args t!j) \<in> supt) \<and> 
         (\<forall>(i,j)\<in>set nstri. i < length(args s) \<and> j < length(args t) \<and> (args s!i, args t!j) \<in> supteq))"
   and check: "check_SCT (\<lambda> st uv. checkEdgRM (computeRuleMap R) (computeRuleMap (reverse_trs R)) Q flag (st,uv)) Gs"
   shows False
proof -
  let ?s = "\<lambda>i. s(Suc i)"
  let ?t = "\<lambda>i. t(Suc i)"
  let ?\<sigma> = "\<lambda>i. \<sigma>(Suc i)"
  let ?rstep = "rstep(set R)"
  let ?supt = "snRel ((supt \<union> ?rstep)^+) ?rstep"
  let ?supteq = "snRel (?rstep^*) ?rstep"
  let ?pairs = "\<lambda>k. ((?s k,?t k),(\<lambda> j. args(?s k)!j \<cdot> ?\<sigma> k))"
  let ?Rs = "?rstep^*"
  let ?R = "{(?pairs k,?pairs(Suc k)) | k. True}"
  let ?conn = "\<lambda>st uv. checkEdgRM (computeRuleMap R) (computeRuleMap(reverse_trs R)) Q flag (st,uv)"
  let ?S = "(?supteq \<union> ?supt)\<^sup>* O ?supt O (?supteq \<union> ?supt)\<^sup>*"
  let ?NS = "(?supteq \<union> ?supt)\<^sup>*"
  have sct_semantics: "sct_semantics ?supt ?supteq"
  proof
    show "SN ?supt"
    proof -
      have "\<forall>(t::('f,'v)term). SN ?supt t"
      proof (rule ccontr)
        assume "\<not>(\<forall>(t::('f,'v)term). SN ?supt t)"
        then obtain t::"('f,'v)term" where "\<not> SN ?supt t" by auto
        then obtain S where "S 0 = t" and "\<forall>i. (S i,S(Suc i)) \<in> ?supt"
          unfolding SN_defs by auto
        hence "\<forall>i. SN ?rstep (S i)" unfolding snRel_def by simp
        hence "SN ?rstep (S 0)" by simp
        hence "SN (supt \<union> ?rstep) t" unfolding `S 0 = t`
          by (rule SN_rstep_seq_imp_SN_supt_union_rstep_seq)

        hence "SN ((supt \<union> ?rstep)^+) t" by (rule SN_step_imp_SN_trancl)
        moreover have "?supt \<subseteq> (supt \<union> ?rstep)^+" by (rule snRel_subset)
        ultimately have "SN ?supt t" by (rule SN_imp_SN_subset)

        with `\<not> SN ?supt t` show False by simp
      qed
      thus ?thesis unfolding SN_defs by simp
    qed

  next

    show "?supt O ?supteq \<subseteq> ?supt"
    proof (rule subsetI2)
      fix s t assume "(s,t) \<in> ?supt O ?supteq"
      then obtain u where fst: "(s,u) \<in> ?supteq" and snd: "(u,t) \<in> ?supt" by auto

      have "(s,u) \<in> ?Rs" and "SN ?rstep s" using fst unfolding snRel_def by auto
      moreover have "(u,t) \<in> (supt \<union> ?rstep)^+" and "SN ?rstep u" using snd
        unfolding snRel_def by auto
      ultimately show "(s,t) \<in> ?supt"
      proof (induct)
        case base thus ?case unfolding snRel_def by simp
      next
        case (step v w)
        from `(v,w) \<in> ?rstep` have "(v,w) \<in> ?rstep^+" by simp
        with trancl_union_right[where r="?rstep" and s="supt"]
          have "(v,w) \<in> (supt \<union> ?rstep)^+" by blast
        with `(w,t) \<in> (supt \<union> ?rstep)^+` have "(v,t) \<in> (supt \<union> ?rstep)^+" by simp
        moreover have "SN ?rstep v" by (rule rtrancl_step_preserves_SN[OF `(s,v) \<in> ?Rs` `SN ?rstep s`])
        ultimately show ?case using step by simp
      qed
    qed
  qed

  from chain have min_t: "\<forall>i. SN ?rstep (t i \<cdot> \<sigma> i)" by (auto simp: minimal_cond_def)

  from chain have inR: "\<forall>k. (t k \<cdot> \<sigma> k,s(Suc k) \<cdot> \<sigma>(Suc k)) \<in> ?rstep^*" by auto

  have "\<forall>i>(0::nat). SN ?rstep (s i \<cdot> \<sigma> i)"
  proof (intro impI allI)
    fix i::nat assume "i > 0"
    then obtain n where "i = Suc n" by (cases i) auto
    from inR have step: "(t(i - 1) \<cdot> \<sigma>(i - 1),s i \<cdot> \<sigma> i) \<in> ?rstep^*" unfolding `i = Suc n` by auto
    from min_t have "SN ?rstep (t(i - 1) \<cdot> \<sigma>(i - 1))" unfolding `i = Suc n` by auto
    from rtrancl_step_preserves_SN[OF step this] show "SN ?rstep (s i \<cdot> \<sigma> i)" .
  qed
  hence min_s: "\<forall>i. SN ?rstep (?s i \<cdot> ?\<sigma> i)" by auto

  have main: "?R \<subseteq> (\<Union>G \<in> set Gs. sct_semantics.steps ?supt ?supteq ?conn G)"
  proof (simp, clarify, simp)
    fix k
    let ?sk = "\<lambda>i. args(?s k)!i \<cdot> ?\<sigma> k"
    let ?tk = "\<lambda>i. args(?t k)!i \<cdot> ?\<sigma> k"
    let ?sk1 = "\<lambda>i. args(?s(Suc k))!i \<cdot> ?\<sigma>(Suc k)"
    from chain have inP: "(?s k,?t k) \<in> P" by auto
    from chain have inR: "(?t k \<cdot> ?\<sigma> k,?s(Suc k) \<cdot> ?\<sigma>(Suc k)) \<in> ?rstep^*" by auto

    have tk_supt: "\<forall>i<length(args(?t k)). (?t k \<cdot> ?\<sigma> k,?tk i) \<in> supt"
    proof (intro impI allI)
      fix i assume i: "i < length(args(?t k))" thus "(?t k \<cdot> ?\<sigma> k,?tk i) \<in> supt"
      proof (cases "?t k")
        case (Var x) with tuple and inP show ?thesis by auto
      next
        case (Fun f ss)
        from i have "(?t k,args(?t k)!i) \<in> supt" unfolding Fun
          using supt.arg[of "args(Fun f ss)!i" "ss" f] by auto
        thus ?thesis by (rule suptp_stable[to_set])
      qed
    qed

    have sk_supt: "\<forall>i<length(args(?s k)). (?s k \<cdot> ?\<sigma> k,?sk i) \<in> supt"
    proof (intro impI allI)
      fix i assume i: "i < length(args(?s k))" thus "(?s k \<cdot> ?\<sigma> k,?sk i) \<in> supt"
      proof (cases "?s k")
        case (Var x) with tuple and inP show ?thesis by auto
      next
        case (Fun f ss)
        from i have "(?s k,args(?s k)!i) \<in> supt" unfolding Fun
          using supt.arg[of "args(Fun f ss)!i" "ss" f] by auto
        thus ?thesis by (rule suptp_stable[to_set])
      qed
    qed

    from chain have sntk: "SN ?rstep (?t k \<cdot> ?\<sigma> k)" by (simp add: minimal_cond_def)
    hence SN_tki: "\<forall>i<length(args(?t k)). SN ?rstep (?tk i)"
      using subterm_preserves_SN[OF sntk] tk_supt by auto

    have SN_sk: "SN ?rstep (?s k \<cdot> ?\<sigma> k)" using min_s by simp
    hence SN_ski: "\<forall>i<length(args(?s k)). SN ?rstep (?sk i)"
      using subterm_preserves_SN[OF SN_sk] sk_supt by auto

    from inP tuple have validDP: "\<not> is_var(?t k) \<and> the(root(?t k)) \<notin> defs (set R)" by auto
    from chain inP tuple have validDP1: "\<not> is_var(?s(Suc k))" by auto
    from chain have steps: "(?t k \<cdot> ?\<sigma> k,?s(Suc k) \<cdot> ?\<sigma>(Suc k)) \<in> ?Rs" by auto
    have stepsNS: "\<forall>j<length(args(?t k)). (?tk j,?sk1 j) \<in> ?NS"
    proof (intro allI impI)
      fix j
      let ?tj = "args(?t k)!j"
      let ?sj = "args(?s(Suc k))!j"
      assume j: "j < length(args(?t k))"

      from validDP obtain f ts where tk: "?t k = Fun f ts \<and> f \<notin> defs (set R)" by (cases "?t k", auto)
      hence f: "f \<notin> defs(set R)" ..
      from validDP1 obtain g ss where sk1: "?s(Suc k) = Fun g ss" by (cases "?s(Suc k)") auto

      let ?tss = "map (\<lambda>t. t \<cdot> ?\<sigma> k) ts"
      let ?sss = "map (\<lambda>s. s \<cdot> ?\<sigma>(Suc k)) ss"

      from steps have "(Fun f ts \<cdot> ?\<sigma> k, Fun g ss \<cdot> ?\<sigma>(Suc k)) \<in> ?Rs" using tk sk1 by auto
      hence steps2: "(Fun f ?tss, Fun g ?sss) \<in> ?Rs" by auto
      from this no_left_vars f have "\<exists> us. length us = length ?tss \<and> Fun g ?sss = Fun f us \<and> (\<forall> i < length ?tss. (?tss ! i, us ! i) \<in> ?Rs)"
        by (rule nondef_root_imp_arg_steps)
      from this obtain us where nearly: "length us = length ?tss \<and> Fun g ?sss = Fun f us \<and> (\<forall> i < length ?tss. (?tss ! i, us ! i) \<in> ?Rs)" ..
      from SN_tki have SN_tkj: "SN ?rstep (?tj \<cdot> ?\<sigma> k)" using j by auto
      from nearly j tk sk1 have "(?tj \<cdot> ?\<sigma> k,?sj \<cdot> ?\<sigma>(Suc k)) \<in> ?Rs" by auto
      with SN_tkj have "(?tj \<cdot> ?\<sigma> k,?sj \<cdot> ?\<sigma>(Suc k)) \<in> snRel ?Rs ?rstep" unfolding snRel_def by simp
      thus "(?tj \<cdot> ?\<sigma> k,?sj \<cdot> ?\<sigma>(Suc k)) \<in> ?NS" unfolding snRel_def by auto
    qed
    from steps have conn: "?conn (?s k,?t k) (?s(Suc k),?t(Suc k))" by simp
    from inP obtain stri nstri where graph: "Scg (?s k,?t k) (?s k,?t k) stri nstri \<in> set Gs \<and> 
         (\<forall>(i,j)\<in>set stri. i < length(args(?s k)) \<and> j < length(args(?t k)) \<and> (args(?s k)!i,args(?t k)!j) \<in> supt) \<and> 
         (\<forall>(i,j)\<in>set nstri. i < length(args(?s k)) \<and> j < length(args(?t k)) \<and> (args(?s k)!i,args(?t k)!j) \<in> supteq)" using graphs by force
    let ?g = "Scg (?s k,?t k) (?s k,?t k) stri nstri"
    from graph have gGs: "?g \<in> set Gs" ..
    let ?check = "\<lambda>g. (((?s k,?t k),?sk),(?s(Suc k),?t(Suc k)),?sk1)
           \<in> sct_semantics.steps ?supt ?supteq ?conn g"
    have "?check ?g"
    proof (simp add: sct_semantics.steps.simps sct_semantics conn, rule conjI)
      {
        fix i j
        assume "(i,j) \<in> set nstri"
        with graph have i: "i < length(args(?s k))" and j: "j < length(args(?t k))"
          and NS: "(args(?s k)!i,args(?t k)!j) \<in> supteq" by auto
        from j stepsNS have astepsNS: "(?tk j,?sk1 j) \<in> ?NS" by auto

        from SN_ski and i have snski: "SN ?rstep (?sk i)" by simp
        from NS supteq_sub_closed have "(?sk i,?tk j) \<in> supteq"
          unfolding sub_closed_def using csI by blast
        hence non_strict: "(?sk i,?tk j) \<in> ?NS" using snski
          unfolding snRel_def supteq_supt_eq_conv by force
        from rtrancl_trans[OF non_strict astepsNS] have "(?sk i,?sk1 j) \<in> ?NS" by auto
      }
      thus "\<forall>ij\<in>set nstri.(\<lambda>(i,j). (?sk i,?sk1 j) \<in> ?NS) ij" by auto
    next
      {
        fix i j
        assume "(i,j) \<in> set stri"
        with graph have i: "i < length (args(?s k))" and j: "j < length(args(?t k))"
          and S: "(args(?s k)!i, args(?t k)!j) \<in> supt" by auto
        from j stepsNS have astepsNS: "(?tk j,?sk1 j) \<in> ?NS" by auto

        from SN_ski and i have snski: "SN ?rstep (?sk i)" by simp
        from S supt_sub_closed have "(?sk i,?tk j) \<in> supt" unfolding sub_closed_def using csI by blast
        hence strict: "(?sk i,?tk j) \<in> ?supt" using snski
          unfolding snRel_def by force
        have "(?sk i,?sk1 j) \<in> ?S" using strict astepsNS by auto
      }
      thus "\<forall>ij\<in>set stri.(\<lambda>(i,j). (?sk i,?sk1 j) \<in> ?S) ij" by auto
    qed
    with gGs show "\<exists>g\<in>set Gs. ?check g" by blast
  qed
  have "SN ?R"
    by (rule sct_semantics.SCT_correctness2[where S = ?supt and NS = ?supteq],rule sct_semantics,rule main,rule check)
  have "\<not> SN ?R" unfolding SN_ars_def SN_elt_def by force
  with `SN ?R` show False by auto
qed

text {* towards automation *}

(* use rulemaps to check tuple and no_left_var condition (as they are precomputed in anyway).
  this might also be useful for subterm-criterion implementation *)
lemma ruleMap_not_None_eq_no_vars_left:
  shows "(computeRuleMap R \<noteq> None) = (\<forall> lr \<in> set R. \<forall> x. fst lr \<noteq> Var x)"
proof (cases "computeRuleMap R")
  case None
  from None computeRuleMapSound2 have "\<exists> (l,r) \<in> set R. root l = None" by auto
  from this obtain l r where "(l,r) \<in> set R \<and> root l = None" by auto
  from this have "(l,r) \<in> set R \<and> \<not> (\<forall> x. fst (l,r) \<noteq> Var x)" by (cases l, auto)
  with None show ?thesis by blast
next
  case (Some rm)
  hence left: "computeRuleMap R \<noteq> None" by auto
  from Some computeRuleMapSound2 have "\<forall> (l,r) \<in> set R. root l \<noteq> None" by force
  hence "\<forall> lr \<in> set R. \<forall> x. fst lr \<noteq> Var x" by auto
  with left show ?thesis by blast
qed

lemma ruleMap_empty_eq_nondef: fixes R :: "('f,'v)trsL"
  assumes "computeRuleMap R = Some rm"
  shows "(rm f = []) = (f \<notin> defs (set R))"
proof -
  from assms computeRuleMapSound have rm: "\<And> f :: 'f. set (rm f) = {(l,r) | l r. (l, r) \<in> set R \<and> root l = Some f}" by force
  show ?thesis
  proof (cases "rm f")
    case Nil
    with rm have "f \<notin> defs (set R)" unfolding defs_def defined_def by force
    with Nil show ?thesis by blast
  next
    case (Cons lr RR)
    hence left: "rm f \<noteq> []" by auto
    from Cons rm[where f = f] have "f \<in> defs (set R)" unfolding defs_def defined_def by (cases lr, force)
    with left show ?thesis by blast
  qed
qed

definition check_supt :: "('f::shows,'v::shows)rule \<Rightarrow> shows monad"
where "check_supt lr \<equiv> checkB (checkSupterm (fst lr) (snd lr))
                              (''the rhs is not a proper subterm of the lhs'' +#+ shows_nl)
                           \<leftarrow> (\<lambda>s. ''Could not orient rule '' +#+ shows_rule lr +@+ shows_nl +@+ s)"

definition check_supteq :: "('f::shows,'v::shows)rule \<Rightarrow> shows monad"
where "check_supteq lr \<equiv> checkB (fst lr = snd lr \<or> checkSupterm (fst lr) (snd lr))
                              (''the rhs is not a subterm of the lhs'' +#+ shows_nl)
                           \<leftarrow> (\<lambda>s. ''Could not orient rule '' +#+ shows_rule lr +@+ shows_nl +@+ s)"

lemma check_supt_sound[simp]:
  assumes "isOK(check_supt r)" shows "r \<in> supt"
proof -
  from assms have "checkSupterm (fst r) (snd r)" unfolding check_supt_def by simp
  thus ?thesis unfolding checkSuptermSound[to_set] by simp
qed

lemma check_supteq_sound[simp]:
  assumes "isOK(check_supteq r)" shows "r \<in> supteq"
proof -
  from assms have "(fst r) = (snd r) \<or> checkSupterm (fst r) (snd r)" unfolding check_supteq_def by simp
  hence "fst r \<unrhd> snd r" unfolding checkSuptermSound by auto
  from this[to_set] show ?thesis by simp
qed

fun check_sct_subterm_proc :: "
  ('f::shows,'v::shows)trsL \<Rightarrow>
  (('f,'v)rule \<times> ((nat\<times>nat)list \<times> (nat\<times>nat)list))list \<Rightarrow>
  ('f,'v)trsL \<Rightarrow> ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow>
  ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> shows monad"
where "check_sct_subterm_proc P Gs R rm rrm = (
  let edge_check = checkEdgRM rm rrm P True in (
    case rm of None \<Rightarrow> error(''TRS has variables as left-hand sides'' +#+ shows_nl)
             | Some rm \<Rightarrow> (
               checkSubset P (map fst Gs)
               \<leftarrow> (\<lambda>s. ''there is no size-change graph for DP '' +#+ shows_rule s +@+ shows_nl)
               \<guillemotright> checkAllM (\<lambda>((s,t),(stri,nstri)). ((
                 checkB (\<not> is_var s \<and> \<not> is_var t \<and> rm(the(root t)) = [])
                        (''roots of DP must be constructors'' +#+ shows_nl)
               \<guillemotright> (let ss = args s in
                  let ts = args t in
                  let n  = length ss in
                  let m  = length ts in (
                    checkAllM (\<lambda>(i,j). checkB (i < n \<and> j < m \<and> isOK(check_supt (ss!i,ts!j)))
                                              (''problem with edge '' +#+ shows i +@+ '' |> '' +#+ shows j +@+ shows_nl)) stri
                  \<guillemotright> checkAllM (\<lambda>(i,j). checkB (i < n \<and> j < m \<and> isOK(check_supteq (ss!i,ts!j)))
                                              (''problem with edge '' +#+ shows i +@+ '' |>= '' +#+ shows j +@+ shows_nl)) nstri
                  ))
               )) \<leftarrow> (\<lambda>m. ''problems with DP '' +#+ shows_rule(s,t) +@+ shows_nl +@+ m)) Gs
             \<guillemotright> checkB (check_SCT (\<lambda>st uv. edge_check(st,uv)) (map (\<lambda>(st,(stri,nstri)). Scg st st stri nstri) Gs))
                      (''size-change analysis failed'' +#+ shows_nl)
             )
    ) \<leftarrow> (\<lambda>s. ''could not apply the size-change processor with subterm-relation for the following reason'' +#+ shows_nl +@+ s)
)"

lemma check_sct_subterm_proc_sound:
  "chain_identifying_onlyDP_proc_impl (\<lambda>P R rm rrm rP. isOK(check_sct_subterm_proc P Gs R rm rrm))"
unfolding chain_identifying_onlyDP_proc_impl_def
proof (intro allI impI)
  fix P R rP s t \<sigma>
  assume check: "isOK(check_sct_subterm_proc P Gs R
                   (computeRuleMap R) (computeRuleMap(reverse_trs R)))"
    and chain: "min_ichain (set P,set R) s t \<sigma>"
  from check obtain rm where rm: "computeRuleMap R = Some rm" by (cases "computeRuleMap R") simp
  let ?Gs = "map (\<lambda>(st,(stri,nstri)). Scg st st stri nstri) Gs"
  let ?conn = "\<lambda>st uv. checkEdgRM (computeRuleMap R) (computeRuleMap(reverse_trs R)) P True (st,uv)"
  from check rm have check1: "set P \<subseteq> fst ` set Gs \<and> check_SCT ?conn ?Gs" by simp
  let ?check2'' = "\<lambda>f g s t stri nstri. \<not> is_var s \<and> \<not> is_var t \<and> rm(the(root t)) = [] \<and>
    (\<forall>(i,j) \<in> set stri. i < length(args s) \<and> j < length(args t) \<and> f(args s!i,args t!j)) \<and>
    (\<forall>(i,j) \<in> set nstri. i < length(args s) \<and> j < length(args t) \<and> g(args s!i,args t!j))"
  let ?check2'  = "?check2'' (\<lambda>r. isOK(check_supt r)) (\<lambda>r. isOK(check_supteq r))"
  let ?check2    = "?check2'' (\<lambda>r. r \<in> supt) (\<lambda>r. r \<in> supteq)"
  from check have "\<forall>((s,t),(stri,nstri)) \<in> set Gs. ?check2' s t stri nstri"
    unfolding rm by (force simp:Let_def)
  hence check2: "\<forall>((s,t),(stri,nstri)) \<in> set Gs. ?check2 s t stri nstri"
    using check_supt_sound check_supteq_sound by fast
  from rm have no_left_vars: "\<forall>lr\<in>set R.\<forall>x. fst lr \<noteq> Var x"
    using ruleMap_not_None_eq_no_vars_left by force
  let ?tuple = "\<lambda>s t. \<not> is_var s \<and> \<not> is_var t \<and> the(root t) \<notin> defs(set R)"
  let ?graph = "\<lambda>s t.\<exists>stri nstri. (
    Scg (s,t) (s,t) stri nstri \<in> set ?Gs \<and>
    (\<forall>(i,j)\<in>set stri. i < length(args s) \<and> j < length(args t) \<and> (args s!i,args t!j) \<in> supt) \<and>
    (\<forall>(i,j)\<in>set nstri. i < length(args s) \<and> j < length(args t) \<and> (args s!i,args t!j) \<in> supteq)
  )"
  {
    fix s t
    assume "(s,t) \<in> set P"
    with check1 obtain stri nstri where mem: "((s,t),(stri,nstri)) \<in> set Gs" by auto
    hence memGs: "Scg (s,t) (s,t) stri nstri \<in> set ?Gs" by force
    from mem check2 have ch: "?check2 s t stri nstri" by auto
    from rm ch ruleMap_empty_eq_nondef have tuple: "?tuple s t" by force
    from ch memGs have "?graph s t" unfolding split_def by auto
    with tuple have "?tuple s t \<and> ?graph s t" by auto
  }
  hence tuple_graph: "\<forall>(s,t)\<in>set P. ?tuple s t \<and> ?graph s t" by auto
  from tuple_graph have tuple: "\<forall>(s,t)\<in>set P. ?tuple s t" by auto
  from tuple_graph have graph: "\<forall>(s,t)\<in>set P. ?graph s t" by auto
  have False by (rule sct_with_subterm[where Q=P and flag=True and Gs = "?Gs"],rule chain,rule no_left_vars,rule tuple,rule graph,simp add: check1)
  thus "\<exists>i. ichain (set rP,set R) (shift s i) (shift t i) (shift \<sigma> i)" by simp
qed

fun check_sct_proc :: "
  (('f::shows,'v::shows)rule \<Rightarrow> shows monad) \<Rightarrow>
  (('f,'v)rule \<Rightarrow> shows monad) \<Rightarrow>
   shows \<Rightarrow>
   ('f,'v)trsL \<Rightarrow> (('f,'v)rule \<times> ((nat \<times> nat)list \<times> (nat \<times> nat)list))list \<Rightarrow> ('f,'v)trsL \<Rightarrow>
  ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow>
  ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> shows monad"
where "check_sct_proc checkSS checkNSS representation P Gs R rm rrm = (
  let edgeCheck = checkEdgRM rm rrm P True in (
    case rm of None \<Rightarrow> error(''TRS has variables as left-hand sides'' +#+ shows_nl)
      | Some RM \<Rightarrow> (
        checkSubset P (map fst Gs)
           \<leftarrow> (\<lambda>s. ''there is no size-change graph for DP '' +#+ shows_rule s)
      \<guillemotright> checkAllM (\<lambda> ((s,t),(stri,nstri)). ((
              checkB (\<not> is_var s \<and> \<not> is_var t \<and> RM (the (root t)) = []) (''roots of DP must be constructors'' +#+ shows_nl)
         \<guillemotright> (let ss = args s in let ts = args t in let n = length ss in let m = length ts in (
                checkAllM (\<lambda> (i,j). checkB (i < n \<and> j < m \<and> isOK(checkSS (ss ! i, ts ! j))) (''problem with edge '' +#+ shows i +@+ '' -S-> '' +#+ shows j)) stri
              \<guillemotright> checkAllM (\<lambda> (i,j). checkB (i < n \<and> j < m \<and> isOK(checkNSS (ss ! i, ts ! j))) (''problem with edge '' +#+ shows i +@+ '' -NS-> '' +#+ shows j)) nstri
              ))
           )) \<leftarrow> (\<lambda>m. ''problems with DP '' +#+ shows_rule(s,t) +@+ shows_nl +@+ m)
        ) Gs
      \<guillemotright> checkAllM checkNSS R
              \<leftarrow> (\<lambda>s. ''problem when orienting rules'' +#+ shows_nl +@+ s)
      \<guillemotright> checkB (check_SCT (\<lambda> st uv. edgeCheck (st,uv)) (map (\<lambda> (st,(stri,nstri)). Scg st st stri nstri) Gs)) (''size-change analysis failed'' +#+ shows_nl)
      ))
  \<leftarrow> (\<lambda>s. ''could not apply the size-change processor with the following '' +#+ shows_nl +@+
         representation +@+ shows_nl +@+ ''for the following reason'' +#+ shows_nl +@+ s)
)"

context redpair_impl
begin

lemma check_sct_proc_sound_1:
   assumes chain: "ichain (set P,set R) s t \<sigma>"
   and check: "isOK(check_sct_proc checkS checkNS repr P Gs R (computeRuleMap R) (computeRuleMap (reverse_trs R)))"
   shows False
proof -
  from check obtain rm where rm: "computeRuleMap R = Some rm" by (cases "computeRuleMap R", auto)
  let ?Gs = "(map (\<lambda> (st,(stri,nstri)). Scg st st stri nstri) Gs)"
  let ?conn = "(\<lambda> st uv. checkEdgRM (computeRuleMap R) (computeRuleMap (reverse_trs R)) P True (st,uv))"
  from check rm have check1: "set P \<subseteq> fst ` set Gs \<and>
     set R \<subseteq> NS \<and>
     check_SCT ?conn ?Gs"
    by (auto simp: checkNSSound)
  let ?check2 = "\<lambda> s t stri nstri. \<not> is_var s \<and> \<not> is_var t \<and> rm (the (root t)) = [] \<and> 
    (\<forall> (i,j) \<in> set stri. i < length (args s) \<and> j < length (args t) \<and> (args s ! i, args t ! j) \<in> S) \<and> 
    (\<forall> (i,j) \<in> set nstri. i < length (args s) \<and> j < length (args t) \<and> (args s ! i, args t ! j) \<in> NS)"
  from check rm have check2: "\<forall> ((s,t),(stri,nstri)) \<in> set Gs. ?check2 s t stri nstri"
    by (force simp: Let_def checkNSSound checkSSound)
  from rm have no_left_vars: "\<forall> lr \<in> set R. \<forall> x. fst lr \<noteq> Var x" using ruleMap_not_None_eq_no_vars_left by force
  let ?tuple = "\<lambda> s t. \<not> is_var s \<and> \<not> is_var t \<and> the (root t) \<notin> defs (set R)"
  let ?graph = "\<lambda> s t. \<exists> stri non_stri. (
         Scg (s,t) (s,t) stri non_stri \<in> set ?Gs \<and>
         (\<forall> (i,j) \<in> set stri. i < length (args s) \<and> j < length (args t) \<and> (args s ! i, args t ! j) \<in> S) \<and> 
         (\<forall> (i,j) \<in> set non_stri. i < length (args s) \<and> j < length (args t) \<and> (args s ! i, args t ! j) \<in> NS))"
  {
    fix s t
    assume "(s,t) \<in> set P"
    with check1 obtain stri nstri where mem: "((s,t),(stri,nstri)) \<in> set Gs" by auto
    hence memGs: "Scg (s,t) (s,t) stri nstri \<in> set ?Gs" by force
    from mem check2 have ch: "?check2 s t stri nstri" by auto
    from rm ch ruleMap_empty_eq_nondef have tuple: "?tuple s t" by force
    from ch memGs have "?graph s t" by blast
    with tuple have "?tuple s t \<and> ?graph s t" by auto
  }
  hence  tuple_graph:  "\<forall> (s,t) \<in> set P. ?tuple s t \<and> ?graph s t" by auto

  from tuple_graph have tuple: "\<forall> (s,t) \<in> set P. ?tuple s t" by blast
  from tuple_graph have graph: "\<forall> (s,t) \<in> set P. ?graph s t" by blast
  show False
    by (rule sct_with_redpair[where Q = P and flag = True], rule chain, rule no_left_vars, rule tuple, simp add: check1, rule graph, auto simp: check1)
qed

lemma check_sct_proc_sound:
  "chain_identifying_onlyDP_proc_impl (\<lambda> P R rm rrm rP. isOK(check_sct_proc checkS checkNS repr P Gs R rm rrm))"
unfolding chain_identifying_onlyDP_proc_impl_def
proof (intro allI, intro impI, intro allI, intro impI)
  fix P R rP s t \<sigma>
  assume check: "isOK(check_sct_proc checkS checkNS repr P Gs R (computeRuleMap R) (computeRuleMap (reverse_trs R)))"
  and chain: "min_ichain (set P, set R) s t \<sigma>"
  hence ichain: "ichain (set P, set R) s t \<sigma>" by auto
  have False by (rule check_sct_proc_sound_1, rule ichain, rule check)
  thus "\<exists> i. ichain (set rP, set R) (shift s i) (shift t i) (shift \<sigma> i)" ..
qed
end

fun check_sct_ur_af_proc :: "
  (('f::shows,'v::shows)rule \<Rightarrow> shows monad) \<Rightarrow>
  (('f,'v)rule \<Rightarrow> shows monad) \<Rightarrow>
   shows \<Rightarrow>
   ('f,'v)trsL \<Rightarrow> (('f,'v)rule \<times> ((nat \<times> nat)list \<times> (nat \<times> nat)list))list \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow>
  ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow>
  ('f,'v)trsL \<Rightarrow> 'f af \<Rightarrow>
  shows monad"
where "check_sct_ur_af_proc checkSS checkNSS representation P Gs R rm rrm ur \<pi> = (
  let edgeCheck = checkEdgRM rm rrm P True in (
    case rm of None \<Rightarrow> error(''TRS has variables as left-hand sides'' +#+ shows_nl)
      | Some RM \<Rightarrow> (
        checkAllM (\<lambda> lr. checkUrClosedTRM_af RM ur \<pi> (snd lr)) ur
           \<leftarrow> (\<lambda>s. ''error when checking closure properties of rhs of usable rules'' +#+ shows_nl +@+ s)
      \<guillemotright> checkSubset P (map fst Gs)
           \<leftarrow> (\<lambda>s. ''there is no size-change graph for DP '' +#+ shows_rule s)
      \<guillemotright> checkAllM (\<lambda>((s,t),(stri,nstri)). ((
              checkB (\<not> is_var s \<and> \<not> is_var t \<and> RM (the (root t)) = []) (''roots of DP must be constructors'' +#+ shows_nl)
         \<guillemotright> checkAllM (\<lambda> i. checkB ( i < length (args s)) (''left-index to large'' +#+ shows i +@+ shows_nl)) (remdups (map fst (List.append stri nstri)))
         \<guillemotright> checkAllM (\<lambda> j. checkB ( j < length (args t) \<and> isOK(checkUrClosedTRM_af RM ur \<pi> (args t ! j))) (''right-index to large or argument violates usable-rules condition'' +#+ shows j)) (remdups (map snd (List.append stri nstri)))
         \<guillemotright> (let ss = args s in let ts = args t in  (
                checkAllM (\<lambda> (i,j). checkB (isOK(checkSS (ss ! i, ts ! j))) (''problem with edge '' +#+ shows i +@+ '' -S-> '' +#+ shows j)) stri
              \<guillemotright> checkAllM (\<lambda> (i,j). checkB (isOK(checkNSS (ss ! i, ts ! j))) (''problem with edge '' +#+ shows i +@+ '' -NS-> '' +#+ shows j)) nstri
              ))
           )) \<leftarrow> (\<lambda> m. ''problems with DP '' +#+ shows_rule(s,t) +@+ shows_nl +@+ m) ) Gs
      \<guillemotright> checkAllM checkNSS ur
              \<leftarrow> (\<lambda>s. ''problem when orienting usable rules'' +#+ shows_nl +@+ s)
      \<guillemotright> checkB (check_SCT (\<lambda> st uv. edgeCheck (st,uv)) (map (\<lambda> (st,(stri,nstri)). Scg st st stri nstri) Gs)) (''size-change analysis failed'' +#+ shows_nl)
      ))
  \<leftarrow> (\<lambda>s. ''could not apply the size-change processor with the following '' +#+ shows_nl +@+
          representation +@+ shows_nl +@+ ''for the following reason'' +#+ shows_nl +@+ s)
)"

context ce_af_redpair_impl
begin

lemma check_sct_ur_af_proc_sound_1:
   assumes chain: "min_ichain (set P,set R) s t \<sigma>"
   and check: "isOK(check_sct_ur_af_proc checkS checkNS repr P Gs R (computeRuleMap R) (computeRuleMap (reverse_trs R)) ur \<pi>)"
   shows False
proof -
  from check obtain rm where rm: "computeRuleMap R = Some rm" by (cases "computeRuleMap R", auto)
  let ?Gs = "(map (\<lambda> (st,(stri,nstri)). Scg st st stri nstri) Gs)"
  let ?conn = "(\<lambda> st uv. checkEdgRM (computeRuleMap R) (computeRuleMap (reverse_trs R)) P True (st,uv))"
  from check rm have check1: "set P \<subseteq> fst ` set Gs \<and>
     set ur \<subseteq> NS \<and>
     urClosed_af (set ur) R \<pi> \<and>
     check_SCT ?conn ?Gs"
    by (auto simp: checkNSSound)
  let ?check2 = "\<lambda> s t stri nstri. \<not> is_var s \<and> \<not> is_var t \<and> rm (the (root t)) = [] \<and> 
    (\<forall> i \<in> fst ` set (stri @ nstri).  i < length (args s)) \<and>
    (\<forall> j \<in> snd ` set (stri @ nstri).  j < length (args t) \<and> urClosedT_af (args t ! j) R (set ur) \<pi>) \<and> 
    (\<forall> (i,j) \<in> set stri. (args s ! i, args t ! j) \<in> S) \<and>
    (\<forall> (i,j) \<in> set nstri. (args s ! i, args t ! j) \<in> NS)"
  from check rm have check2: "\<forall> ((s,t),(stri,nstri)) \<in> set Gs. ?check2 s t stri nstri"
    by (force simp: Let_def checkNSSound checkSSound simp del: set_append map_append)


  from rm have no_left_vars: "\<forall> lr \<in> set R. \<forall> x. fst lr \<noteq> Var x" using ruleMap_not_None_eq_no_vars_left by force
  let ?tuple = "\<lambda> s t. \<not> is_var s \<and> \<not> is_var t \<and> the (root t) \<notin> defs (set R)"
  let ?pgraph = "\<lambda> s t stri non_stri. (
         Scg (s,t) (s,t) stri non_stri \<in> set ?Gs \<and>
         (\<forall> (i,j) \<in> set stri. i < length (args s) \<and> j < length (args t) \<and> urClosedT_af (args t ! j) R (set ur) \<pi> \<and> (args s ! i, args t ! j) \<in> S) \<and> 
         (\<forall> (i,j) \<in> set non_stri. i < length (args s) \<and> j < length (args t) \<and> urClosedT_af (args t ! j) R (set ur) \<pi> \<and> (args s ! i, args t ! j) \<in> NS))"
  let ?graph = "\<lambda> s t. \<exists> stri non_stri. ?pgraph s t stri non_stri"
  {
    fix s t
    assume "(s,t) \<in> set P"
    with check1 obtain stri nstri where mem: "((s,t),(stri,nstri)) \<in> set Gs" by auto
    hence memGs: "Scg (s,t) (s,t) stri nstri \<in> set ?Gs" by force
    from mem check2 have ch: "?check2 s t stri nstri" by auto
    from ch have stri_ur: "\<forall> (i,j) \<in> set stri. i < length (args s) \<and> j < length (args t) \<and> urClosedT_af (args t ! j) R (set ur) \<pi>" by force
    from ch have nstri_ur: "\<forall> (i,j) \<in> set nstri. i < length (args s) \<and> j < length (args t) \<and> urClosedT_af (args t ! j) R (set ur) \<pi>" by force
    from rm ch ruleMap_empty_eq_nondef have tuple: "?tuple s t" by force
    from ch memGs stri_ur nstri_ur have "?pgraph s t stri nstri" by auto
    hence "?graph s t" by blast
    with tuple have "?tuple s t \<and> ?graph s t" by auto
  }
  hence  tuple_graph:  "\<forall> (s,t) \<in> set P. ?tuple s t \<and> ?graph s t" by auto

  from tuple_graph have tuple: "\<forall> (s,t) \<in> set P. ?tuple s t" by blast
  from tuple_graph have graph: "\<forall> (s,t) \<in> set P. ?graph s t" by blast
  show False
    by (rule sct_ur_af_redpair[where Q = P and flag = True and ur = "set ur"], rule chain, rule no_left_vars, rule tuple, simp add: check1, simp add: check1, rule graph, auto simp: check1)
qed

lemma check_sct_ur_af_proc_sound:
  "chain_identifying_onlyDP_proc_impl (\<lambda> P R rm rrm rP. isOK(check_sct_ur_af_proc checkS checkNS repr P Gs R rm rrm ur \<pi>))"
unfolding chain_identifying_onlyDP_proc_impl_def
proof (intro allI, intro impI, intro allI, intro impI)
  fix P R rP s t \<sigma>
  assume check: "isOK(check_sct_ur_af_proc checkS checkNS repr P Gs R (computeRuleMap R) (computeRuleMap (reverse_trs R)) ur \<pi>)"
  and chain: "min_ichain (set P, set R) s t \<sigma>"
  have False by (rule check_sct_ur_af_proc_sound_1, rule chain, rule check)
  thus "\<exists> i. ichain (set rP, set R) (shift s i) (shift t i) (shift \<sigma> i)" ..
qed
end

fun create_sct_ur_af_proc_comp :: "('a::shows)poly_components \<Rightarrow> ('f::shows,'a)lpoly_interL \<Rightarrow> ('f,'v::shows)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
  ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f,'v)trsL \<Rightarrow> (('f,'v)rule \<times> ((nat \<times> nat)list \<times> (nat \<times> nat)list))list \<Rightarrow>
        shows monad"
where "create_sct_ur_af_proc_comp c I P R rm rrm ur graphs = (
  let pi = to_lpoly_inter (get_comp_1 c) I in (
  check_polo_coeffs_comp c I \<guillemotright>
  check_sct_ur_af_proc 
           (check_polo_s_comp c pi)
           (check_polo_ns_comp c pi)
           (''polynomial order ('' +#+ get_comp_description c +#+ '')'' +#+ shows_nl +@+
            create_lpoly_repr c I +@+ shows_nl)
           P graphs R rm rrm ur
           (create_afI (get_comp_0 c) I)
))"

context poly_order_carrier
begin
lemma sct_ur_af_proc_comp_ci : 
  "chain_identifying_onlyDP_proc_impl (\<lambda> P R rm rrm rP. isOK(create_sct_ur_af_proc_comp carrier_components I P R rm rrm ur graphs))"
unfolding "chain_identifying_onlyDP_proc_impl_def"
proof (intro allI, intro impI)
  fix P R rP
  let ?c = carrier_components
  let ?pi = "to_lpoly_inter one I"
  assume check: "isOK(create_sct_ur_af_proc_comp ?c I P R (computeRuleMap R) (computeRuleMap (reverse_trs R)) ur graphs)"
  hence check1: "isOK(check_polo_coeffs_comp ?c I)" by (auto simp: Let_def)
  let ?af = "create_af I"
  from check have accept: "isOK(check_sct_ur_af_proc (check_polo_s_comp ?c ?pi) (check_polo_ns_comp ?c ?pi) id P graphs R (computeRuleMap R) (computeRuleMap (reverse_trs R)) ur ?af)" (is "isOK(?proc)") 
    by (auto simp: Let_def) 
  have "chain_identifying_onlyDP_proc_impl (\<lambda> P R rm rrm rP. isOK(check_sct_ur_af_proc (check_polo_s_comp ?c ?pi) (check_polo_ns_comp ?c ?pi) id P graphs R rm rrm ur ?af))"
    by (rule ce_af_redpair_impl.check_sct_ur_af_proc_sound, rule conditional_inter_ce_af_redpair_impl, rule check1)
  thus "\<forall> s t \<sigma>. min_ichain (set P, set R) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (set rP, set R) (shift s i) (shift t i) (shift \<sigma> i))"
    unfolding chain_identifying_onlyDP_proc_impl_def
    using accept by best
qed
end


end