(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
header {* Util *}
theory Util imports Main Either Rational begin
text {* Some auxiliary lemmas that do not fit elsewhere. *}


subsection {* Properties of @{const take} and @{const drop} *}

lemma nth_append_take:
  assumes "i \<le> length xs" shows "(take i xs @ y#ys)!i = y"
proof -
  from assms have a: "length(take i xs) = i" by simp
  have "(take i xs @ y#ys)!(length(take i xs)) = y" by (rule nth_append_length)
  thus ?thesis unfolding a .
qed

lemma nth_append_take_is_nth_conv:
  assumes "i < j" and "j \<le> length xs" shows "(take j xs @ ys)!i = xs!i"
proof -
  from assms have "i < length(take j xs)" by simp
  hence "(take j xs @ ys)!i = take j xs ! i" unfolding nth_append by simp
  thus ?thesis unfolding nth_take[OF assms(1)] .
qed

lemma nth_append_drop_is_nth_conv:
  assumes "j < i" and "j \<le> length xs" and "i \<le> length xs"
  shows "(take j xs @ y # drop (Suc j) xs)!i = xs!i"
proof -
  from assms have j: "length(take j xs) = j" by auto
  from `j < i` obtain n where "Suc(j + n) = i" using less_imp_Suc_add by auto
  with assms have i: "i = length(take j xs) + Suc n" by auto
  have len: "Suc j + n \<le> length xs" using assms i by auto
  have "(take j xs @ y # drop (Suc j) xs)!i =
    (y # drop (Suc j) xs)!(i - length(take j xs))" unfolding nth_append i by auto
  also have "\<dots> = (y # drop (Suc j) xs)!(Suc n)" unfolding i by simp
  also have "\<dots> = (drop (Suc j) xs)!n" by simp
  finally show ?thesis unfolding nth_drop[OF len] using i j by simp
qed

lemma nth_append_take_drop_is_nth_conv: 
 assumes "i \<le> length xs" and "j \<le> length xs" and "i \<noteq> j" 
 shows "(take j xs @ y # drop (Suc j) xs)!i = xs!i"
proof -
 from assms have "i < j \<or> i > j" by auto
 thus ?thesis
 proof
  assume "i < j" thus ?thesis using assms by (auto simp: nth_append_take_is_nth_conv)
 next
  assume "j < i" thus ?thesis using assms by (auto simp: nth_append_drop_is_nth_conv)
 qed
qed
  
lemma take_drop_imp_nth: "\<lbrakk>take i ss @ x # drop (Suc i) ss = ss\<rbrakk> \<Longrightarrow> x = ss!i"
proof (induct ss arbitrary: i)
 case Nil thus ?case by auto
next
 case (Cons s ss)
 from `take i (s#ss) @ x # drop (Suc i) (s#ss) = (s#ss)` show ?case
 proof (induct i)
  case 0 thus ?case by auto
 next
  case (Suc i)
  from Cons have IH: "take i ss @ x # drop (Suc i) ss = ss \<Longrightarrow> x = ss!i" by auto
  from Suc have "take i ss @ x # drop (Suc i) ss = ss" by auto
  with IH show ?case by auto
 qed
qed


subsection {* Induction Rules *}

text {* Induction over a finite set of natural numbers. *}
lemma bound_nat_induct[consumes 1]:
  assumes "n \<in> {l..u}" and "P l" and "\<And>n. \<lbrakk>P n;n \<in> {l..<u}\<rbrakk> \<Longrightarrow> P(Suc n)"
  shows "P n"
using assms proof (induct n rule: nat.induct)
 case zero thus ?case by simp
next
 case (Suc n) thus ?case
 proof (cases "Suc n = l")
  case True with `P l` show ?thesis by simp
 next
  case False with Suc show ?thesis by auto
 qed
qed


subsection {* Functions on Lists *}

text {* A tail recursive variant for @{term "concat(map f xs)"}. *}
primrec rev_append :: "'a list \<Rightarrow> 'a list \<Rightarrow> 'a list"
where "rev_append [] ys = ys"
    | "rev_append (x#xs) ys = rev_append xs (x#ys)"

text {* @{const rev_append} is characterized through @{const rev} and @{const append}. *}
lemma rev_append_conv[simp]: "rev_append xs ys = rev xs @ ys"
  by (induct xs arbitrary: ys) auto
declare rev_append.simps[simp del]

primrec flat_map_aux :: "('a \<Rightarrow> 'b list) \<Rightarrow> 'a list \<Rightarrow> 'b list \<Rightarrow> 'b list"
where "flat_map_aux f [] ys = rev ys"
    | "flat_map_aux f (x#xs) ys = flat_map_aux f xs (rev_append (f x) ys)"

lemma flat_map_aux_conv: "flat_map_aux f xs ys = rev ys @ concat(map f xs)"
  by (induct xs arbitrary: ys) auto

definition flat_map where "flat_map f xs \<equiv> flat_map_aux f xs []"

text {* {const flat_map_aux} is no longer needed. *}
hide const flat_map_aux

lemma flat_map_concat_map_conv[simp]: "flat_map f xs = concat(map f xs)"
  unfolding flat_map_def flat_map_aux_conv by simp

text {* For code-generation, replace algebraically nice functions by
efficient and/or tail recursive ones. *}
lemma [code inline]: "concat(map f xs) = flat_map f xs"
  by (rule flat_map_concat_map_conv[symmetric])

lemma flat_map_cong[fundef_cong]:
  fixes f g::"'a \<Rightarrow> 'b list"
  assumes "xs = ys" and "\<And>(x::'a). x \<in> set ys \<Longrightarrow> f x = g x"
  shows "flat_map f xs = flat_map g ys"
using map_cong[OF assms] by simp

text {* Check whether there is an element in @{text xs} that satisfies the
predicate @{text p}. *}
primrec any :: "('a \<Rightarrow> bool) \<Rightarrow> 'a list \<Rightarrow> bool"
where "any _ [] = False"
    | "any p (x#xs) = (p x \<or> any p xs)"

lemma any_mem_conv: "any f xs = (\<exists>x. x mem xs \<and> f x)" by (induct xs, auto)
lemma any_set_conv: "any f xs = (\<exists>x \<in> set xs. f x)" by (auto simp: any_mem_conv mem_iff)

lemma any_nth_conv: "any f xs = (\<exists> i < length xs. f (xs ! i))"
  unfolding any_mem_conv mem_iff set_conv_nth by auto

lemma any_cong[fundef_cong] : fixes xs ys f g 
  assumes xs: "xs = ys" and set: "\<And>x. x \<in> set ys \<Longrightarrow> f x = g x" 
  shows "any f xs = any g ys" 
proof -
  from xs have "any f xs = any f ys" by simp
  also have "\<dots> = (\<exists> x. x mem ys \<and> f x)" by (rule any_mem_conv)
  also have "\<dots> = (\<exists> x. x mem ys \<and> g x)" using mem_iff set by best
  also have "\<dots> = any g ys" using any_mem_conv by auto
  finally show ?thesis .
qed

text {* Check whether all elements of the list @{text xs} satisfy the
predicate @{text p}. *}
primrec forall :: "('a \<Rightarrow> bool) \<Rightarrow> 'a list \<Rightarrow> bool"
where "forall p [] = True"
    | "forall p (x#xs) = (p x \<and> forall p xs)"

lemma forall_any_conv:
  "forall p xs = (\<not>(any (\<lambda>x. \<not> p x) xs))" by (induct xs) auto

lemma forall_set_conv[simp] : "forall f xs = (\<forall>x\<in>set xs. f x)"
  by (auto simp: forall_any_conv any_mem_conv mem_iff)

lemma forall_cong[fundef_cong] : fixes xs ys f g 
  assumes xs: "xs = ys" and set: "\<And>x. x \<in> set ys \<Longrightarrow> f x = g x" 
  shows "forall f xs = forall g ys" 
proof -
  from xs have "forall f xs = forall f ys" by simp
  also have "\<dots> = (\<forall> x \<in> set ys. f x)" by (rule forall_set_conv)
  also have "\<dots> = (\<forall> x \<in> set ys. g x)" using set by best
  also have "\<dots> = forall g ys" using forall_set_conv by auto
  finally show ?thesis .
qed

declare forall.simps[simp del]

primrec find :: "('a \<Rightarrow> bool) \<Rightarrow> 'a list \<Rightarrow> 'a option"
where "find _ [] = None"
    | "find p (x#xs) = (if p x then Some x else find p xs)"

lemma find_NoneE: 
  assumes "find f xs = None" shows "\<not>(\<exists>x. x mem xs \<and> f x)"
using assms proof (induct xs)
  case Nil show ?case by simp
next
  case (Cons x xs)
  hence "\<not>(f x)" unfolding find.simps by auto
  with Cons show ?case by auto
qed

lemma find_NoneI:
  assumes "(\<not>(\<exists>x. x mem xs \<and> f x))" shows "find f xs = None"
using assms proof (induct xs)
 case Nil
 show ?case by simp
next
 case (Cons x xs)
 assume "\<not>(\<exists>y. y mem (x # xs) \<and> f y)" hence "\<not> (f x)" by auto 
 with Cons show ?case by auto
qed

lemma not_mem_find_None_conv: "(\<not>(\<exists>x. x mem xs \<and> f x)) = (find f xs = None)"
 using find_NoneI find_NoneE by blast

lemma find_SomeE: 
  assumes "find p xs = Some x"
  shows "\<exists>i<length xs. p(xs!i) \<and> x = (xs!i) \<and> (\<forall>j<i. \<not> p(xs!j))"
using assms proof (induct xs)
  case Nil thus ?case by simp
next
  case (Cons y ys)
  show ?case
  proof (cases "p y")
    case True with Cons show ?thesis by auto
  next
    case False with Cons have "find p ys = Some x" by simp
    with Cons have "\<exists>i<length ys. p(ys!i) \<and> x = ys!i \<and> (\<forall>j<i. \<not> p(ys!j))" by simp
    then obtain i where "i < length ys" and "p(ys!i)" and "x = ys!i"
      and "\<forall>j<i. \<not> p(ys!j)" by auto
    from `i < length ys` have "Suc i < length(y#ys)" by simp
    moreover from `p(ys!i)` have "p((y#ys)!Suc i)" by simp
    moreover from `x = ys!i` have "x = (y#ys)!(Suc i)" by simp
    moreover have "\<forall>j<Suc i. \<not> p((y#ys)!j)"
    proof (intro allI impI)
      fix j assume "j < Suc i" thus "\<not>p((y#ys)!j)" using `\<forall>j<i. \<not>p(ys!j)` False
        by (induct j) auto
    qed
    ultimately show ?thesis by best
  qed
qed

lemma find_SomeI:
  assumes "\<exists>i<length xs. f(xs!i) \<and> x = (xs!i) \<and> (\<forall>j<i. \<not> f(xs!j))"
  shows "find f xs = Some x"
using assms proof (induct xs)
  case Nil thus ?case by simp
next
  case (Cons y ys)
  with `\<exists> i < length (y # ys). f ((y # ys) ! i) \<and> (x =  ((y # ys) ! i)) \<and> (\<forall> j < i. \<not> f ((y # ys) ! j))`
  obtain i where "i < length (y # ys)" and "f ((y # ys) ! i)" 
           and "x = ((y # ys) ! i)" and "\<forall> j < i. \<not> f ((y # ys) ! j)" by auto
  show ?case
  proof (cases i, simp)
   case 0
    with `f ((y # ys) ! i)` have "f y" by simp
    with  `x = (y # ys) ! i` 0 have "y = x" by simp
    with `f y` show " (f y \<longrightarrow> y = x) \<and> (\<not> f y \<longrightarrow> find f ys = Some x)" by auto
  next
   case (Suc k)
    from Suc `\<forall> j < i. \<not> f ((y # ys) ! j)` have "\<not> f y" by auto
    hence rec:"find f (y # ys) = find f ys" by auto
    from Suc `i < length (y # ys)` have k1:"k < length ys" by simp
    from Suc `f ((y # ys) ! i)` have k2:"f (ys ! k)" by simp
    from Suc `x = (y # ys) ! i` have k3:"x = (ys ! k)" by simp
    from Suc `\<forall> j < i. \<not> f ((y # ys) ! j)` have "\<forall> j < k. \<not> f (ys ! j)" by auto
    with k1 k2 k3 have "\<exists>i<length ys. f (ys ! i) \<and> x = ys ! i \<and> (\<forall>j<i. \<not> f (ys ! j))" by auto
    with Cons have "find f ys = Some x" by auto
    with rec show "find f (y#ys) = Some x" by simp
  qed
qed

lemma ex_find_Some_conv:
 "(\<exists>i<length xs. f(xs!i) \<and> x = (xs!i) \<and> (\<forall>j<i. \<not> f(xs!j))) = (find f xs = Some x)"
by (rule iffI,rule find_SomeI,assumption,rule find_SomeE,assumption)

lemma find_cong[fundef_cong]:
  assumes xs: "xs = ys" and set: "\<And>x. x \<in> set ys \<Longrightarrow> f x = g x" 
  shows "find f xs = find g ys"
proof (cases "find f xs")
  case None
  hence "find f ys = None" using xs by simp
  hence "\<not>(\<exists> x. x mem ys \<and> f x)" using find_NoneE by auto
  hence "\<not>(\<exists> x. x mem ys \<and> g x)" using mem_iff set by best   
  hence "find g ys = None" using find_NoneI by auto
  thus ?thesis using None by auto
next
  case (Some x) 
  hence "find f ys = Some x" using xs by simp
  hence "\<exists>i<length ys. f(ys!i) \<and> (x = (ys!i)) \<and> (\<forall>j<i. \<not> f (ys!j))" by (rule find_SomeE[where ?xs=ys])
  hence "\<exists>i<length ys. g(ys!i) \<and> (x = (ys!i)) \<and> (\<forall>j<i. \<not> g (ys!j))" using mem_iff set by auto
  hence "find g ys = Some x" by (rule find_SomeI) 
  thus ?thesis using Some by auto
qed

lemma partition_filter_conv[simp]:
  "partition f xs = (filter f xs,filter (Not o f) xs)"
  unfolding partition_filter2[symmetric]
  unfolding partition_filter1[symmetric] by simp

declare partition.simps[simp del]

primrec list_union :: "'a list \<Rightarrow> 'a list \<Rightarrow> 'a list"
where "list_union [] ys = ys"
   |  "list_union (x#xs) ys = (let zs = list_union xs ys in if x mem zs then zs else x # zs)"

lemma list_union_sound[simp] : "set(list_union xs ys) = set xs \<union> set ys"
proof (induct xs, simp)
  case (Cons x xs)
  thus ?case by (cases "x mem (unionI xs ys)", auto simp: mem_iff Let_def)
qed

declare list_union.simps[simp del]

abbreviation map_filter :: "('a \<Rightarrow> bool) \<Rightarrow> ('a \<Rightarrow> 'b) \<Rightarrow> 'a list \<Rightarrow> 'b list"
where "map_filter f g xs \<equiv> foldr (\<lambda> x rest. if f x then g x # rest else rest) xs []"

fun lookup :: "'a \<Rightarrow> ('a \<times> 'b)list \<Rightarrow> 'b option"
where "lookup _ [] = None"
   |  "lookup a ((a',b) # xs) = (if a = a' then Some b else lookup a xs)"

lemma lookup_complete: assumes "lookup a ps = None"
  shows "\<not>(\<exists>b. (a,b) \<in> set ps)"
using assms
proof (induct ps, simp)
  case (Cons ab xs)
  show ?case
  proof (cases ab)
    case (Pair a' b)
    with Cons show ?thesis by (cases "a = a'", auto)
  qed
qed

lemma lookup_sound: assumes "lookup a ps = Some b"
  shows "(a,b) \<in> set ps"
using assms
proof (induct ps, simp)
  case (Cons ab xs)
  show ?case
  proof (cases ab)
    case (Pair a' b)
    with Cons show ?thesis by (cases "a = a'", auto)
  qed
qed

subsection {* Lists and Sets *}

lemma length_remdups_card_conv: "length(remdups xs) = card(set xs)"
proof -
  have xs: "concat[xs] = xs" by simp
  from length_remdups_concat[of "[xs]"] show ?thesis unfolding xs by simp
qed

lemma set_foldr_remdups_set_map_conv[simp]:
  "set(foldr (\<lambda>x xs. remdups(f x@xs)) xs []) = \<Union>set(map (set o f) xs)"
  by (induct xs) auto

lemma zip_same: "((a,b) \<in> set (zip xs xs)) = (a \<in> set xs \<and> a = b)"
  by (induct xs, auto)

subsection {* normalization of rational numbers *}
definition normalize :: "rat \<Rightarrow> int \<times> int" where
  "normalize x \<equiv> THE pair. x = Fract (fst pair) (snd pair) \<and> snd pair > 0 \<and> zgcd (fst pair) (snd pair) = 1"

declare normalize_def[code del]

lemma normalize_code[code,simp]: "normalize (Fract a b) = (
  if b > 0 then (let g = zgcd a b in (a div g, b div g))
  else if b = 0 then (0,1) 
  else (let g = - zgcd a b in (a div g, b div g)))"
proof - 
  let ?cond = "\<lambda> r p. r = Fract (fst p) (snd p) \<and> snd p > 0 \<and> zgcd (fst p) (snd p) = 1"
  show ?thesis 
  proof (cases "b = 0")
    case True
    thus ?thesis 
    proof (simp add: normalize_def)
      show "(THE pair. ?cond (Fract a 0) pair) = (0,1)"
      proof
	show "?cond (Fract a 0) (0,1)"
	  by (simp add: rat_number_collapse)
      next
	fix pair
	assume cond: "?cond (Fract a 0) pair"
	show "pair = (0,1)"
	proof (cases pair)
	  case (Pair den num)
	  with cond have num: "num > 0" by auto
	  with Pair cond have den: "den = 0" by (simp add: eq_rat)
	  show ?thesis
	  proof (cases "num = 1", simp add: Pair den)
	    case False
	    with num have gr: "num > 1" by auto
	    with den have "zgcd den num = num" by auto
	    with Pair cond False gr 
	    show ?thesis by auto
	  qed
	qed
      qed
    qed
  next
    {
      fix a b :: int
      assume b: "b > 0"
      hence b0: "b \<noteq> 0" and "b \<ge> 0" by auto
      let ?g = "zgcd a b"
      from b0 have g0: "?g \<noteq> 0" using zgcd0 by auto
      with zgcd_pos[of a b] have gp: "?g > 0" by arith
      from zgcd_zdvd1[of a b] obtain a' where a': "a = ?g * a'" unfolding dvd_def by auto
      from zgcd_zdvd2[of a b] obtain b' where b': "b = ?g * b'" unfolding dvd_def by auto
      hence b'2: "b' * ?g = b" by (simp add: ring_simps)
      with b0 have b'0: "b' \<noteq> 0" by auto
      from b b' zero_less_mult_iff[of ?g b'] have "0 < ?g \<and> 0 < b' \<or> ?g < 0 \<and> b' < 0" by auto
      with gp have b'p: "b' > 0" by auto
      have "\<exists> l. b' = l + 1 \<and> l \<ge> 0" by (rule exI[of _ "b' - 1"], simp add: b'p)
      (* for whatever reason the following crashes when using the command line!
          hence "\<exists> l. b' = l + (1 :: int) \<and> l \<ge> 0" by arith *)
      then obtain l where l: "l + 1 = b'" and lp: "l \<ge> 0" by auto
      have "?g \<le> 0 * ?g + 1 * ?g" by simp
      also have "\<dots> \<le> l * ?g + 1 * ?g" using gp mult_right_mono[OF lp, of ?g] by auto
      also have "\<dots> = (l + 1) * ?g" by (simp add: ring_distribs)
      finally have gs: "?g \<le> b" by (simp add: l b'2)
      have "normalize (Fract a b) = (a div ?g, b div ?g)"
      proof (simp add: normalize_def)
	show "(THE pair. ?cond (Fract a b) pair) = (a div ?g, b div ?g)"
	proof
	  have "1 = b div b" using b0 by auto
	  also have "\<dots> \<le> b div ?g" by (rule zdiv_mono2[OF `b \<ge> 0` gp gs])
	  finally have div0: "b div ?g > 0" by simp
	  show "?cond (Fract a b) (a div ?g, b div ?g)"
	    by (simp add: b0 Fract_norm div_zgcd_relprime div0)
	next
	  fix pair
	  assume cond: "?cond (Fract a b) pair"
	  show "pair = (a div ?g, b div ?g)"
	  proof (cases pair)
	    case (Pair den num)
	    with cond have num: "num > 0" and num0: "num \<noteq> 0" and gcd: "zgcd den num = 1" by auto
	    obtain g where g: "g = ?g" by auto
	    with gp have gg0: "g \<noteq> 0" by auto
	    from cond Pair eq_rat(1)[OF b0 num0] have eq: "a * num = den * b" by auto
	    hence "a' * g * num = den * g * b'" using a'[simplified g[symmetric]] b'[simplified g[symmetric]] by simp
	    hence "a' * num * g = b' * den * g" by (simp add: ring_simps)
	    hence eq2: "a' * num = b' * den" using gg0 by auto
	    have "a div ?g = ?g * a' div ?g" using a' by force
	    hence adiv: "a div ?g = a'" using g0 by auto
	    have "b div ?g = ?g * b' div ?g" using b' by force
	    hence bdiv: "b div ?g = b'" using g0 by auto
	    from div_zgcd_relprime[of a b] b0 have "zgcd (a div ?g) (b div ?g) = 1" by auto
	    with adiv bdiv have gcd2: "zgcd a' b' = 1" by auto
	    from gcd have gcd3: "zgcd num den = 1" by (simp add: zgcd_commute[of den num])
	    from gcd2 have gcd4: "zgcd b' a' = 1" by (simp add: zgcd_commute[of a' b'])
	    have one: "num dvd b'" by (rule zrelprime_dvd_mult[OF gcd3], simp add: dvd_def, rule exI[of _ a'], simp add: eq2 ring_simps)
	    have two: "b' dvd num" by (rule zrelprime_dvd_mult[OF gcd4], simp add: dvd_def, rule exI[of _ den], simp add: eq2 ring_simps)
	    from one two obtain k k' where k: "num = b' * k" and k': "b' = num * k'" unfolding dvd_def by auto
	    hence "num = num * (k * k')" by (simp add: ring_simps)
	    with num0 have prod: "k * k' = 1" by auto
	    from zero_less_mult_iff[of b' k] b'p num k have kp: "k > 0" by auto
	    from prod pos_zmult_eq_1_iff[OF kp, of k'] have "k = 1" by auto
	    with k have numb': "num = b'" by auto
	    with eq2 b'0 have "a' = den" by auto
	    with numb' adiv bdiv Pair show ?thesis by simp
	  qed
	qed
      qed
    } note main = this
    assume "b \<noteq> 0"
    hence "b > 0 \<or> b < 0" by arith
    thus ?thesis
    proof
      assume b: "b > 0"
      thus ?thesis by (simp add: Let_def main[OF b])
    next
      assume b: "b < 0"
      have "\<exists> c. b = -c" by (rule exI[of _ "-b"], simp)
      with b obtain c where bc: "b = - c" and c: "c > 0" by auto
      have id: "Fract a b = Fract (-a) (-b)" by simp
      from c show ?thesis by (simp only: id, simp add: bc, simp add: main[OF c, of "-a"], simp add: zdiv_zminus2)
    qed
  qed
qed

subsection {* Miscellaneous *}

text {* Check whether a list of sets forms a partition, i.e.,
whether the sets are pairwise disjoint. *}
definition is_partition :: "('a set)list \<Rightarrow> bool"
where "is_partition cs = (\<forall>j<length cs. \<forall>i<j. cs!i \<inter> cs!j = {})"

end
