(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Xml imports ProofChecker 
begin

datatype xml = XML string "(string \<times> string) list" "xml list" "string option" (* name, attributes, children, text *)
datatype xmldoc = XMLDOC "string" xml (* header, body *)

text {* An alternative induction scheme for xml. *}
lemma
 fixes P :: "xml \<Rightarrow> bool"
 assumes "\<And>n a cs t. (\<And>c. c \<in> set cs \<Longrightarrow> P c) \<Longrightarrow> P(XML n a cs t)"
 shows xml_induct[case_names XML, induct type: xml]: "P xml" and "\<And>xml. xml \<in> set xmls \<Longrightarrow> P xml"
using assms by (induct xml and xmls) auto

text {* Resolve problem with recursive dependencies in generated code. *}
lemma [code]:
  shows "HOL.eq (XML name1 attrs1 children1 text1) (XML name2 attrs2 children2 text2) \<longleftrightarrow>
    HOL.eq name1 name2 \<and> HOL.eq attrs1 attrs2 \<and> list_all2 HOL.eq children1 children2 \<and> HOL.eq text1 text2"
  by (simp_all add: eq list_all2_eq [symmetric])

abbreviation tab :: "char" where "tab \<equiv> Char Nibble0 Nibble9"
abbreviation carriage_return :: "char" where "carriage_return \<equiv> Char Nibble0 NibbleD"
definition wspace :: "string" where "wspace = newline # tab # carriage_return # '' ''"
definition dquote :: "string" where "dquote = ''\"''"
definition letters :: "string" where "letters = ''abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_0123456789&;:''"

fun trim :: "string \<Rightarrow> string" 
  where "trim (c # s) = (if c mem wspace then trim s else c # s)"
     |  "trim [] = []"

fun splitterFun :: "(char \<Rightarrow> bool) \<Rightarrow> string \<Rightarrow> string \<times> string"
  where "splitterFun allowed (c # s) = 
  (if allowed c then (let (l,r) = splitterFun allowed s in (c # l, r)) else ([], c # s))"
     |  "splitterFun _ [] = ([],[])"

fun splitter :: "string \<Rightarrow> string \<Rightarrow> string \<times> string"
  where "splitter allowed s = splitterFun (\<lambda> c. c mem allowed) s"

types 'a parser = "string \<Rightarrow> (string,'a \<times> string)emonad"
types 'a endParser = "string \<Rightarrow> (string,'a)emonad"

definition eofErr :: "(string,'a)emonad" 
  where "eofErr = error ''unexpected end of input''"

fun parseError :: "string \<Rightarrow> string \<Rightarrow> (string,'a)emonad"
  where "parseError s allowed = error (''expected symbols '' @ allowed @ '' but found '' @ take 30 s)"

fun parseName :: "string parser" 
  where "parseName s = (let (l,r) = splitter letters s in (if l = [] then parseError s letters else return (l,trim r)))" 

fun parseText :: "string option parser" 
  where "parseText s = (let (l,r) = splitterFun (\<lambda> c. [c] \<noteq> ''<'') s in (if trim l = [] then return (None,r) else return (Some (List.rev (trim (List.rev (trim l)))),r)))"

fun consume2 :: "string list \<Rightarrow> string list \<Rightarrow> string parser"
  where "consume2 [] allowed s = parseError s (concat (map (\<lambda> s. s @ '','') allowed))"
     |  "consume2 (t # as) allowed s = (if (map snd (zip t s)) = t then return (t,trim (List.drop (length t) s)) else consume2 as allowed s)"

fun consume :: "string list \<Rightarrow> string parser"
  where "consume allowed s = consume2 allowed allowed s"

fun scanUpto :: "string \<Rightarrow> string parser"
  where "scanUpto end [] = error (''did not find required token '' @ end)"
      | "scanUpto end (c # s) = (if map snd (zip end (c # s)) = end 
             then return (take (length end) (c # s), trim (List.drop (length end) (c # s)))
             else scanUpto end s \<guillemotright>= (\<lambda> (res,rest). return (c # res, rest)))"

fun parseAttribute :: "string parser"
  where "parseAttribute s = (consume [dquote] s \<guillemotright>= 
            (\<lambda> (_,s1). let (n,s2) = splitterFun (\<lambda> c. [c] \<noteq> dquote) s1 in 
                  consume [dquote] s2 \<guillemotright>= (\<lambda> (_,s3). return (n,s3))))" 

(*
lemma bindCong [fundef_cong] : fixes m1 m2 :: "('e,'a)emonad"
  assumes "m1 = m2" and "n1 = n2"
  shows "(m1 \<guillemotright>= n1) = (m2 \<guillemotright>= n2)"
using assms
 by (unfold "bind_def", auto)
*)

function parseAttributes :: "(string \<times> string) list parser"
  where "parseAttributes [] = return ([],[])"
      | "parseAttributes (c # s) = (if c mem ''/>'' then return ([],c # s) else 
                        parseName (c # s) \<guillemotright>= 
                   (\<lambda> (n,s1). consume [''=''] s1 \<guillemotright>=
                   (\<lambda> (_,s2). parseAttribute s2 \<guillemotright>= 
                   (\<lambda> (a,s3). parseAttributes s3 \<guillemotright>=
                   (\<lambda> (nas,s4). return ((n,a) # nas,s4))))))" 
by (pat_completeness, auto)

function parseXMLList :: "xml list parser"
where "parseXMLList s = (if (s = [] \<or> take 2 s = ''</'') then return ([],s) else 
 (consume [''<''] s 
  \<guillemotright>= (\<lambda> (_,s1). parseName s1 
  \<guillemotright>= (\<lambda> (n,s2). parseAttributes s2
  \<guillemotright>= (\<lambda> (nas,s3). consume [''/>'',''>''] s3
  \<guillemotright>= (\<lambda> (clo,s4). if clo = ''/>'' then parseXMLList s4 \<guillemotright>= (\<lambda> (li, s5). return (XML n nas [] None # li, s5)) else 
    parseText s4 
  \<guillemotright>= (\<lambda> (t,s5). parseXMLList s5
  \<guillemotright>= (\<lambda> (chi,s6). consume [''</'' @ n @ ''>''] s6
  \<guillemotright>= (\<lambda> (_,s7). parseXMLList s7 
  \<guillemotright>= (\<lambda> (li,s8). return (XML n nas chi t # li, s8) 
  ))))))))))" 
by (pat_completeness, auto)

fun parseXML :: "xml parser"
where "parseXML s = (consume [''<''] s 
  \<guillemotright>= (\<lambda> (_,s1). parseName s1 
  \<guillemotright>= (\<lambda> (n,s2). parseAttributes s2
  \<guillemotright>= (\<lambda> (nas,s3). consume [''/>'',''>''] s3
  \<guillemotright>= (\<lambda> (clo,s4). if clo = ''/>'' then return (XML n nas [] None, trim s4) else 
    parseText s4 
  \<guillemotright>= (\<lambda> (t,s5). parseXMLList s5
  \<guillemotright>= (\<lambda> (chi,s6). consume [''</'' @ n @ ''>''] s6
  \<guillemotright>= (\<lambda> (_,s7). return (XML n nas chi t, s7)
  ))))))))"

function stripXMLHeader :: "string parser"
where "stripXMLHeader s = (if take 2 (trim s) = ''<?'' 
          then (scanUpto ''?>'' s
                \<guillemotright>= (\<lambda> (header1,s1). stripXMLHeader s1 \<guillemotright>= (\<lambda> (header2, s2). return (header1 @ header2, trim s2))))
          else return ([],trim s))"
by (pat_completeness, auto)

fun remove_comments_help :: "bool \<Rightarrow> string \<Rightarrow> string"
where "remove_comments_help False (c # cs) = (if (c = CHR ''<'' \<and> take 3 cs = ''!--'') 
   then remove_comments_help True (tl cs)
   else c # remove_comments_help False cs)"
 | "remove_comments_help True (c # cs) = (if (c = CHR ''-'' \<and> take 2 cs = ''->'') 
      then remove_comments_help False (drop 2 cs)
      else remove_comments_help True cs)"
 | "remove_comments_help _ [] = []"

definition remove_comments :: "string \<Rightarrow> string"
where "remove_comments \<equiv> remove_comments_help False"

fun parseXMLDoc :: "xmldoc endParser"
where "parseXMLDoc s = debug ''0'' ''parsing txt to xml'' (stripXMLHeader (remove_comments s)
     \<guillemotright>= (\<lambda> (header, rest). parseXML rest \<guillemotright>= (\<lambda> (xml, end). if end = [] then return (XMLDOC header xml) else 
         error (''there is text after root element: '' @ newline # end))))"

(* and pretty printer *)

instantiation xml :: "shows" begin
  definition shows_attr :: "string\<times>string \<Rightarrow> shows"
  where "shows_attr av = (shows(fst av) +@+ ''='' +#+ shows_string(dquote @ (snd av) @ dquote))"
  definition shows_attrs :: "(string\<times>string)list \<Rightarrow> shows"
  where "shows_attrs as = shows_map (\<lambda>a. '' '' +#+ shows_attr a) as"
  fun shows_XML_indent :: "string \<Rightarrow> nat \<Rightarrow> xml \<Rightarrow> shows"
  where "shows_XML_indent ind i (XML n a c t) = (
    ind +#+ ''<'' +#+ shows n +@+ shows_attrs a +@+ (
      if c = [] \<and> t = None then shows_string ''/>'' else (
        ''>'' +#+ (
          if c \<noteq> [] then (
            shows_map (shows_XML_indent (ind @ replicate i (CHR '' '')) i) c +@+ shows_string ind
          ) else id
        ) +@+ (if t = None then id else shows_string(the t)) +@+
          ''</'' +#+ shows n +@+ shows_string ''>''
      )
    )
  )"
  definition shows_xml :: "xml \<Rightarrow> shows" where "shows_xml xml \<equiv> shows_XML_indent [newline] 2 xml"
  definition shows_list_xml :: "xml list \<Rightarrow> shows" where "shows_list_xml xs \<equiv> shows_list_aux shows xs"
  
  lemma shows_attr_assoc: "(s +#+ shows_attr av) r @ t = (s +#+ shows_attr av) (r@t)"
    unfolding shows_attr_def by (cases av) (auto simp: show_defs o_def dquote_def)
  lemma shows_attr_assoc_elt: "\<forall>av\<in>set avs. (s +#+ shows_attr av) r @ t = (s +#+ shows_attr av) (r@t)"
    using shows_attr_assoc by best
  lemma shows_attrs_assoc: "shows_attrs as r @ s = shows_attrs as (r@s)"
    unfolding shows_attrs_def shows_map_assoc[OF shows_attr_assoc_elt[of as "'' ''"]] by simp
  lemma assoc_xml': "shows_XML_indent ind i xml r @ s = shows_XML_indent ind i xml (r@s)"
  proof (induct xml arbitrary: ind i r s)
    let ?elt = "shows_XML_indent"
    case (XML n a cs t ind i)
    hence elt: "\<And>r s.\<forall>c\<in>set cs. ?elt (ind @ replicate i CHR '' '') i c r @ s = ?elt (ind @ replicate i CHR '' '') i c (r@s)" by auto
    have IH: "\<And>r s. shows_map (?elt (ind @ replicate i CHR '' '') i) cs r @ s = shows_map (?elt (ind @ replicate i CHR '' '') i) cs (r@s)"
      by (rule shows_map_assoc[OF elt])
    show ?case
    proof (cases "cs = []")
      case True show ?thesis
      proof (cases t)
        case None
        with True have empty: "cs = [] \<and> t = None" by simp
        show ?thesis by (auto simp: shows_attrs_assoc if_P[OF empty])
      next
        case (Some ts)
        hence t: "\<not>(cs = [] \<and> t = None)" and some_t: "\<not> t = None" by auto
        from True have empty: "\<not> cs \<noteq> []" by simp
        show ?thesis by (auto simp: if_not_P[OF t] if_not_P[OF empty] if_not_P[OF some_t] shows_attrs_assoc)
      qed
    next
      case False
      hence ne: "\<not>(cs = [] \<and> t = None)" by simp
      show ?thesis
      proof (cases t)
        case None show ?thesis by (auto simp: if_not_P[OF ne] if_P[OF False] if_P[OF None] shows_attrs_assoc IH)
      next
        case (Some ts)
        hence some_t: "\<not> t = None" by simp
        show ?thesis by (auto simp: if_not_P[OF ne] if_P[OF False] if_not_P[OF some_t] shows_attrs_assoc IH)
      qed
    qed
  qed

  lemma assoc_xml: "shows (xml::xml) r @ s = shows xml (r@s)" unfolding shows_xml_def by (rule assoc_xml')
  lemma assoc_xml_elt: "\<forall>(x::xml)\<in>set xs. shows x r @ s = shows x (r@s)" using assoc_xml by auto

  instance by
    (intro_classes,rule assoc_xml,unfold shows_list_xml_def)
    (rule shows_list_aux_assoc[OF assoc_xml_elt])
end

instantiation xmldoc :: "shows" begin
  fun shows_xmldoc where "shows_xmldoc(XMLDOC h x) = (h +#+ shows_nl +@+ shows x)"
  definition shows_list_xmldoc where "shows_list_xmldoc(xs::xmldoc list) \<equiv> shows_list_aux shows xs"

  lemma assoc_xmldoc: "shows (x::xmldoc) r @ s = shows x (r@s)" by (cases x) auto
  lemma assoc_xmldoc_elt: "\<forall>(x::xmldoc)\<in>set xs. shows x r @ s = shows x (r@s)" using assoc_xmldoc by auto

  instance by
    (intro_classes,rule assoc_xmldoc,unfold shows_list_xmldoc_def)
    (rule shows_list_aux_assoc[OF assoc_xmldoc_elt])
end

(* xml to ...  *)

fun mapMonad :: "(xml \<Rightarrow> ('e,'a)emonad) \<Rightarrow> xml list \<Rightarrow> ('e,'a list)emonad"
where "mapMonad _ [] = return []"
    | "mapMonad f (xml # rest) = (case f xml of Left e \<Rightarrow> error e | Right a \<Rightarrow> (case mapMonad f rest of Left e \<Rightarrow> error e | Right as \<Rightarrow> return (a # as)))"

lemma mapMonadCong [fundef_cong] : fixes f1 f2 :: "xml \<Rightarrow> ('e,'a)emonad"
  assumes "f1 = f2" and "l1 = l2"
  shows "(mapMonad f1 l1) = (mapMonad f2 l2)"
using assms by auto


fun xmlGetName :: "xml \<Rightarrow> string" 
where "xmlGetName (XML name _ _ _) = name"


definition string2bool :: "bool endParser"
where "string2bool s \<equiv> (
  if s = ''true'' then return True else (
    if s = ''false'' then return False else error(''cannot convert '' @ s @ '' into Boolean'')
  )
)"

fun xml2text :: "string \<Rightarrow> xml \<Rightarrow> (string,string)emonad"
where "xml2text reqName (XML name atts cs text) = 
  (if name = reqName \<and> atts = [] \<and> cs = [] \<and> text \<noteq> None \<and> the text \<noteq> [] then return (the text) else 
  error (concat [''could not extract text for '',reqName,'' from  '',[newline],show(XML name atts cs text)]))"

fun xml2bool :: "string \<Rightarrow> xml \<Rightarrow> (string,bool)emonad"
where "xml2bool reqName x = (xml2text reqName x \<guillemotright>= string2bool)"

fun xml0elements :: "string \<Rightarrow> 'a \<Rightarrow> xml \<Rightarrow> (string,'a)emonad"
where "xml0elements reqName combi (XML name atts cs text) = 
  (if name = reqName \<and> atts = [] \<and> cs = [] \<and> text = None then return combi 
     else error (concat [''could not transform the following xml element (expected '',reqName,'')'',[newline],show(XML name atts cs text)]))
    " 

fun xml1element :: "string \<Rightarrow> (xml \<Rightarrow> (string,'a)emonad) \<Rightarrow> ('a \<Rightarrow> 'b) \<Rightarrow> xml \<Rightarrow> (string,'b)emonad"
where "xml1element reqName p1 combi (XML name atts cs text) = 
  (if name = reqName \<and> atts = [] \<and> length cs = 1 \<and> text = None then
    (p1 (cs ! 0) \<guillemotright>= (return \<circ> combi))
  else error (concat [''could not transform the following xml element (expected '',reqName,'')'',[newline],show(XML name atts cs text)]))
    " 

fun xml2elements :: "string \<Rightarrow> (xml \<Rightarrow> (string,'a)emonad) \<Rightarrow> (xml \<Rightarrow> (string,'b)emonad) \<Rightarrow> ('a \<Rightarrow> 'b \<Rightarrow> 'c) \<Rightarrow> xml \<Rightarrow> (string,'c)emonad"
where "xml2elements reqName p1 p2 combi (XML name atts cs text) = 
  (if name = reqName \<and> atts = [] \<and> length cs = 2 \<and> text = None then
    (p1 (cs ! 0) \<guillemotright>= (\<lambda> a. p2 (cs ! 1) \<guillemotright>= (return \<circ> combi a)))
  else error (concat [''could not transform the following xml element (expected '',reqName,'')'',[newline],show(XML name atts cs text)]))
    "

fun xml3elements :: "string \<Rightarrow> (xml \<Rightarrow> (string,'a)emonad) \<Rightarrow> (xml \<Rightarrow> (string,'b)emonad) \<Rightarrow> (xml \<Rightarrow> (string,'c)emonad) \<Rightarrow> ('a \<Rightarrow> 'b \<Rightarrow> 'c \<Rightarrow> 'd) \<Rightarrow> xml \<Rightarrow> (string,'d)emonad"
where "xml3elements reqName p1 p2 p3 combi (XML name atts cs text) = 
  (if name = reqName \<and> atts = [] \<and> length cs = 3 \<and> text = None then
    (p1 (cs ! 0) \<guillemotright>= 
    (\<lambda> a. p2 (cs ! 1) \<guillemotright>= 
    (\<lambda> b. p3 (cs ! 2) 
    \<guillemotright>= (return \<circ> combi a b))))
  else error (concat [''could not transform the following xml element (expected '',reqName,'')'',[newline],show(XML name atts cs text)]))
    "

fun xml4elements :: "string \<Rightarrow> (xml \<Rightarrow> (string,'a)emonad) \<Rightarrow> (xml \<Rightarrow> (string,'b)emonad) \<Rightarrow> (xml \<Rightarrow> (string,'c)emonad) \<Rightarrow> (xml \<Rightarrow> (string,'d)emonad) \<Rightarrow> ('a \<Rightarrow> 'b \<Rightarrow> 'c \<Rightarrow> 'd \<Rightarrow> 'e) \<Rightarrow> xml \<Rightarrow> (string,'e)emonad"
where "xml4elements reqName p1 p2 p3 p4 combi (XML name atts cs text) = 
  (if name = reqName \<and> atts = [] \<and> length cs = 4 \<and> text = None then
    (p1 (cs ! 0) \<guillemotright>= 
    (\<lambda> a. p2 (cs ! 1) \<guillemotright>= 
    (\<lambda> b. p3 (cs ! 2) \<guillemotright>= 
    (\<lambda> c. p4 (cs ! 3) 
    \<guillemotright>= (return \<circ> combi a b c)))))
  else error (concat [''could not transform the following xml element (expected '',reqName,'')'',[newline],show(XML name atts cs text)]))
    "

fun xml5elements :: "string \<Rightarrow> (xml \<Rightarrow> (string,'a)emonad) \<Rightarrow> (xml \<Rightarrow> (string,'b)emonad) \<Rightarrow> (xml \<Rightarrow> (string,'c)emonad) \<Rightarrow> (xml \<Rightarrow> (string,'d)emonad) \<Rightarrow> (xml \<Rightarrow> (string,'e)emonad) \<Rightarrow> ('a \<Rightarrow> 'b \<Rightarrow> 'c \<Rightarrow> 'd \<Rightarrow> 'e \<Rightarrow> 'f) \<Rightarrow> xml \<Rightarrow> (string,'f)emonad"
where "xml5elements reqName p1 p2 p3 p4 p5 combi (XML name atts cs text) = 
  (if name = reqName \<and> atts = [] \<and> length cs = 5 \<and> text = None then
    (p1 (cs ! 0) \<guillemotright>= 
    (\<lambda> a. p2 (cs ! 1) \<guillemotright>= 
    (\<lambda> b. p3 (cs ! 2) \<guillemotright>= 
    (\<lambda> c. p4 (cs ! 3) \<guillemotright>= 
    (\<lambda> d. p5 (cs ! 4) \<guillemotright>= 
    (return \<circ> combi a b c d))))))
  else error (concat [''could not transform the following xml element (expected '',reqName,'')'',[newline],show(XML name atts cs text)]))
    "
 
fun xml1to2elements :: "string \<Rightarrow> (xml \<Rightarrow> (string,'a)emonad) \<Rightarrow> (xml \<Rightarrow> (string,'b)emonad) \<Rightarrow> ('a \<Rightarrow> 'b option \<Rightarrow> 'c) \<Rightarrow> xml \<Rightarrow> (string,'c)emonad"
where "xml1to2elements reqName p1 p2 combi (XML name atts cs text) = 
  (if name = reqName \<and> atts = [] \<and> length cs \<ge> 1 \<and> length cs \<le> 2 \<and> text = None then
    (p1 (cs ! 0) \<guillemotright>= (\<lambda> a. if (length cs = 2) then (p2 (cs ! 1) \<guillemotright>= (\<lambda> b. return (combi a (Some b)))) else return (combi a None)))
  else error (concat [''could not transform the following xml element (expected '',reqName,'')'',[newline],show(XML name atts cs text)]))
    " 

(* parses the first element, the second argument must be a boolean. If the boolean is true, then the last element is parsed additionally *)
fun xml2ndChoice :: "string \<Rightarrow> (xml \<Rightarrow> (string,'a)emonad) \<Rightarrow> string \<Rightarrow> (xml \<Rightarrow> (string,'b)emonad) \<Rightarrow> ('a \<Rightarrow> 'b option \<Rightarrow> 'c) \<Rightarrow> xml \<Rightarrow> (string,'c)emonad"
where "xml2ndChoice reqName p1 choiceName p2 combi (XML name atts cs text) = 
  (if name = reqName \<and> atts = [] \<and> length cs \<ge> 2 \<and> text = None then
    (p1 (cs ! 0) \<guillemotright>= (\<lambda> a. xml2bool choiceName (cs ! 1) \<guillemotright>= (\<lambda> choice. if choice then (p2 (cs ! (length cs - 1)) \<guillemotright>= (\<lambda> b. return (combi a (Some b)))) else return (combi a None))))
  else error (concat [''could not transform the following xml element (expected '',reqName,'')'',[newline],show(XML name atts cs text)]))
    " 

fun xml2to3elements :: "string \<Rightarrow> (xml \<Rightarrow> (string,'a)emonad) \<Rightarrow> (xml \<Rightarrow> (string,'b)emonad) \<Rightarrow> (xml \<Rightarrow> (string,'c)emonad) \<Rightarrow> ('a \<Rightarrow> 'b \<Rightarrow> 'c option \<Rightarrow> 'd) \<Rightarrow> xml \<Rightarrow> (string,'d)emonad"
where "xml2to3elements reqName p1 p2 p3 combi (XML name atts cs text) = 
  (if name = reqName \<and> atts = [] \<and> length cs \<ge> 2 \<and> length cs \<le> 3 \<and> text = None then
    (p1 (cs ! 0) \<guillemotright>= (\<lambda> a. p2 (cs ! 1) \<guillemotright>= (\<lambda> b. if (length cs = 3) then (p3 (cs ! 2) \<guillemotright>= (\<lambda> c. return (combi a b (Some c)))) else return (combi a b None))))
  else error (concat [''could not transform the following xml element (expected '',reqName,'')'',[newline],show (XML name atts cs text)]))
    " 

fun xmlManyElements :: "string \<Rightarrow> (xml \<Rightarrow> (string,'a)emonad) \<Rightarrow> ('a list \<Rightarrow> 'b) \<Rightarrow> xml \<Rightarrow> (string,'b)emonad"
where "xmlManyElements reqName p combi (XML name atts cs text) = 
  (if name = reqName \<and> atts = [] \<and> text = None then
     (mapMonad p cs \<guillemotright>= (return \<circ> combi))
   else error (concat [''could not transform the following xml element (expected '',reqName,'')'',[newline],show(XML name atts cs text)]))"

fun xml1CmanyElements :: "string \<Rightarrow> (xml \<Rightarrow> (string,'a)emonad) \<Rightarrow> ('a \<Rightarrow> xml \<Rightarrow> (string,'b)emonad) \<Rightarrow> ('a \<Rightarrow> 'b list \<Rightarrow> 'c) \<Rightarrow> xml \<Rightarrow> (string,'c)emonad"
where "xml1CmanyElements reqName p1 p2 combi (XML name atts cs text) = 
  (if name = reqName \<and> atts = [] \<and> text = None \<and> cs \<noteq> [] then
     (p1 (hd cs) \<guillemotright>= (\<lambda> a. mapMonad (p2 a) (tl cs) \<guillemotright>= (return \<circ> combi a)))
   else error (concat [''could not transform the following xml element (expected '',reqName,'')'',[newline],show(XML name atts cs text)]))"

definition xml1manyElements :: "string \<Rightarrow> (xml \<Rightarrow> (string,'a)emonad) \<Rightarrow> (xml \<Rightarrow> (string,'b)emonad) \<Rightarrow> ('a \<Rightarrow> 'b list \<Rightarrow> 'c) \<Rightarrow> xml \<Rightarrow> (string,'c)emonad"
where "xml1manyElements reqName p1 p2 = xml1CmanyElements reqName p1 (\<lambda> a. p2)"

fun xml2manyElements :: "string \<Rightarrow> (xml \<Rightarrow> (string,'a)emonad) \<Rightarrow> (xml \<Rightarrow> (string,'b)emonad) \<Rightarrow> (xml \<Rightarrow> (string,'c)emonad) \<Rightarrow> ('a \<Rightarrow> 'b \<Rightarrow> 'c list \<Rightarrow> 'd) \<Rightarrow> xml \<Rightarrow> (string,'d)emonad"
where "xml2manyElements reqName p1 p2 p3 combi (XML name atts cs text) = 
  (if name = reqName \<and> atts = [] \<and> text = None \<and> length cs \<ge> 2 then
     (p1 (cs ! 0) \<guillemotright>= (\<lambda> a. p2 (cs ! 1) \<guillemotright>= (\<lambda> b. mapMonad p3 (tl (tl cs)) \<guillemotright>= (return \<circ> combi a b))))
   else error (concat [''could not transform the following xml element (expected '',reqName,'')'',[newline],show(XML name atts cs text)]))"

fun xmlMany2Elements :: "string \<Rightarrow> (xml \<Rightarrow> (string,'a)emonad) \<Rightarrow> (xml \<Rightarrow> (string,'b)emonad) \<Rightarrow> (xml \<Rightarrow> (string,'c)emonad) \<Rightarrow> ('a list \<Rightarrow> 'b \<Rightarrow> 'c \<Rightarrow> 'd) \<Rightarrow> xml \<Rightarrow> (string,'d)emonad"
where "xmlMany2Elements reqName p1 p2 p3 combi (XML name atts cs text) = (let ds = List.rev cs in 
  (if name = reqName \<and> atts = [] \<and> text = None \<and> length cs \<ge> 2 then
     (mapMonad p1 (List.rev (tl (tl ds))) \<guillemotright>= (\<lambda> a. p2 (ds ! 1) \<guillemotright>= (\<lambda> b. p3 (ds ! 0) \<guillemotright>= (return \<circ> combi a b))))
   else error (concat [''could not transform the following xml element (expected '',reqName,'')'',[newline],show(XML name atts cs text)])))"


fun xmlOptions :: "(string \<times> (xml \<Rightarrow> (string,'a)emonad)) list \<Rightarrow> xml \<Rightarrow> (string,'a)emonad"
where "xmlOptions ps x = (case lookup (xmlGetName x) ps of
          None \<Rightarrow> error (concat [''expected one of: '', concat (map (\<lambda> p. fst p @ '' '') ps), [newline], ''but found'', [newline], show x])
        | Some p \<Rightarrow> p x)"

fun xmlChange :: "(xml \<Rightarrow> (string,'a)emonad) \<Rightarrow> ('a \<Rightarrow> 'b) \<Rightarrow> (xml \<Rightarrow> (string,'b)emonad)"
where "xmlChange p f x = (p x \<guillemotright>= (\<lambda> a. return (f a)))"



fun digit2nat :: "char \<Rightarrow> (string,nat)emonad"
where "digit2nat x = (if x = CHR ''0'' then return 0 else
                     (if x = CHR ''1'' then return 1 else
                     (if x = CHR ''2'' then return 2 else
                     (if x = CHR ''3'' then return 3 else
                     (if x = CHR ''4'' then return 4 else
                     (if x = CHR ''5'' then return 5 else
                     (if x = CHR ''6'' then return 6 else
                     (if x = CHR ''7'' then return 7 else
                     (if x = CHR ''8'' then return 8 else
                     (if x = CHR ''9'' then return 9 else 
                       error ([x] @ '' is not a digit'')))))))))))"

fun string2natHelp :: "nat \<Rightarrow> nat endParser" 
where "string2natHelp n [] = return n"
    | "string2natHelp n (d # s) = (
      digit2nat d \<guillemotright>= (\<lambda> m. string2natHelp (10 * n + m) s))"

definition string2nat :: "nat endParser"
where "string2nat s = (if s = [] then error ''cannot convert empty string into number'' else string2natHelp 0 s)"

fun nat2int :: "nat \<Rightarrow> int"
where "nat2int 0 = 0"
   |  "nat2int (Suc n) = 1+nat2int n"

fun string2int :: "int endParser"
where "string2int s = (case s of [] \<Rightarrow> error ''cannot convert empty string into number''
                           | h # t \<Rightarrow> let (pos,ss) = (if h = CHR ''-'' then (False,t) else (True,s))
                               in string2nat ss \<guillemotright>= (\<lambda> n. if pos \<or> n = 0 then return (nat2int n) else return (0 - nat2int n)))"


fun xml2int :: "string \<Rightarrow> xml \<Rightarrow> (string,int)emonad"
where "xml2int reqName x = (xml2text reqName x \<guillemotright>= string2int)"

fun xml2nat :: "string \<Rightarrow> xml \<Rightarrow> (string,nat)emonad"
where "xml2nat reqName x = (xml2text reqName x \<guillemotright>= string2nat)"

definition xml2rat :: "xml \<Rightarrow> (string,rat)emonad"
where "xml2rat \<equiv> xmlOptions [(''integer'', xmlChange (xml2int ''integer'') (\<lambda> x. Fract x 1)),
        (''rational'', xml2elements ''rational'' (xml2int ''numerator'') (xml2int ''denominator'') Fract)]"

fun xml2var :: "xml \<Rightarrow> (string,('a,string)term)emonad"
where "xml2var x = (xml2text ''var'' x \<guillemotright>= (return \<circ> Var))"


function xml2term :: "(xml \<Rightarrow> (string,'a)emonad) \<Rightarrow> xml \<Rightarrow> (string,('a,string)term)emonad"
where "xml2term xml2name x = 
  xmlOptions [(''var'',xml2var), (''funapp'', xml1manyElements ''funapp'' xml2name (xml1element ''arg'' (xml2term xml2name) id) Fun)] x"  
by (pat_completeness, auto)

function xml2context :: "(xml \<Rightarrow> (string,'a)emonad) \<Rightarrow> xml \<Rightarrow> (string,('a,string)ctxt)emonad"
where "xml2context xml2name x = 
  xmlOptions [
       (''box'', xml0elements ''box'' Hole),  
       (''funContext'', xml4elements ''funContext'' 
          xml2name 
          (xmlManyElements ''before'' (xml2term xml2name) id)
          (xml2context xml2name)
          (xmlManyElements ''after'' (xml2term xml2name) id)
          More)
     ] x"  
by (pat_completeness, auto)

definition xml2flat_contexts :: "(xml \<Rightarrow> (string,'a)emonad) \<Rightarrow> xml \<Rightarrow> (string,('a,string)ctxt list)emonad"
where "xml2flat_contexts xml2name = xmlManyElements ''flatContexts'' (xml2context xml2name) id"

definition xml2substitution :: "(xml \<Rightarrow> (string,'a)emonad) \<Rightarrow> xml \<Rightarrow> (string,('a,string)subL)emonad"
where "xml2substitution xml2name = 
   xmlManyElements ''substitution''
      (xml2elements ''substEntry'' 
         xml2var
         (xml2term xml2name)
         Pair)
      id"

primrec list2position :: "nat list \<Rightarrow> pos"
where "list2position [] = \<epsilon>"
    | "list2position (n # ns) = n <# list2position ns"

definition xml2position :: "xml \<Rightarrow> (string,nat)emonad"
where "xml2position \<equiv> xmlChange (xml2nat ''position'') (\<lambda> n. n - 1)"

definition xml2pos :: "xml \<Rightarrow> (string,pos)emonad"
where "xml2pos \<equiv> xmlManyElements ''positionInTerm'' xml2position list2position"

fun xml2sp :: "(xml \<Rightarrow> (string,'a)emonad) \<Rightarrow> xml \<Rightarrow> (string,'a spL)emonad"
where "xml2sp xml2name =
  xmlManyElements ''argumentFilter''
    (xml3elements ''argumentFilterEntry''
      xml2name
      (xml2nat ''arity'')
      (xml2nat ''collapsing'')
      (\<lambda> f a n. Pair f (n - 1)))
    SimpProj"

definition xml2rule :: "(xml \<Rightarrow> (string,'a)emonad) \<Rightarrow> xml \<Rightarrow> (string,('a,string)rule)emonad"
where "xml2rule xml2name = xml2elements ''rule'' (xml1element ''lhs'' (xml2term xml2name) id) (xml1element ''rhs'' (xml2term xml2name) id) Pair" 

definition xml2rules :: "(xml \<Rightarrow> (string,'a)emonad) \<Rightarrow> xml \<Rightarrow> (string,('a,string)trsL)emonad"
where "xml2rules xml2name = xmlManyElements ''rules'' (xml2rule xml2name) id"


definition xml2funcsym :: "xml \<Rightarrow> (string,string \<times> nat)emonad"
where "xml2funcsym = xml2elements ''funcsym'' (xml2text ''name'') (xml2nat ''arity'') Pair" 

definition xml2signature :: "xml \<Rightarrow> (string,(string \<times> nat)list)emonad"
where "xml2signature = xmlManyElements ''signature'' xml2funcsym id"

types labelType = "nat list" 
types lsymbol = "(string,labelType)lab"
types llsymbol = "(lsymbol shp,labelType)lab"

definition xml2label :: "(xml \<Rightarrow> (string,'a)emonad) \<Rightarrow> xml \<Rightarrow> (string,(labelType,'a list)either)emonad"
where "xml2label xml2name \<equiv> xmlOptions [
  (''numberLabel'', xmlManyElements ''numberLabel'' (xml2nat ''number'') Left),
  (''symbolLabel'', xmlManyElements ''symbolLabel'' xml2name Right)
  ]" 

function xml2plainName :: "xml \<Rightarrow> (string,lsymbol)emonad"
where "xml2plainName x = xmlOptions [
  (''name'',xmlChange (xml2text ''name'') UnLab),
  (''labeledSymbol'', xml2elements ''labeledSymbol'' xml2plainName (xml2label xml2plainName)
    (\<lambda>f l. case l of Left ls  \<Rightarrow> Lab f ls
                   | Right ls \<Rightarrow> FunLab f ls))
] x"
by pat_completeness auto

definition xml2sharpName :: "xml \<Rightarrow> (string,lsymbol shp)emonad"
where "xml2sharpName x = (if xmlGetName x = ''sharp'' then
  xml1element ''sharp'' xml2plainName Sharp x
     else
  xmlChange xml2plainName Plain x
  )"

fun xml2labSharpName :: "nat \<Rightarrow> xml \<Rightarrow> (string,llsymbol)emonad"
where "xml2labSharpName 0 = xmlChange xml2sharpName UnLab"
   |  "xml2labSharpName (Suc n) = xml2elements ''labeledSymbol'' (xml2labSharpName n) (xml2label (xml2labSharpName n))
        (\<lambda>f l. case l of Left ls  \<Rightarrow> Lab f ls
                       | Right ls \<Rightarrow> FunLab f ls)"

fun fitLength :: "'a \<Rightarrow> nat \<Rightarrow> 'a list \<Rightarrow> 'a list"
where "fitLength _ 0 _ = []"
    | "fitLength a n [] = replicate n a"
    | "fitLength a (Suc n) (b # as) = b # (fitLength a n as)"

function xml2arith_fun :: "xml \<Rightarrow> (string,arithFun)emonad"
where "xml2arith_fun xml = xml1element ''arithFunction'' (xmlOptions [ 
  (''natural'', xmlChange (xml2nat ''natural'') Const),
  (''variable'', xmlChange (xml2nat ''variable'') (\<lambda> n :: nat. Arg (n - 1))),
  (''sum'', xmlManyElements ''sum'' xml2arith_fun Sum),
  (''product'', xmlManyElements ''product'' xml2arith_fun Prod),
  (''min'', xmlManyElements ''min'' xml2arith_fun Min),
  (''max'', xmlManyElements ''max'' xml2arith_fun Max)
  ]) id  xml"
by (pat_completeness, auto)

definition xml2sl_inter :: "(xml \<Rightarrow> (string,('f,'l)lab)emonad) \<Rightarrow> xml \<Rightarrow> (string,('f,'l)sl_inter option)emonad"
where "xml2sl_inter xml2name \<equiv> xml1element ''model'' (xmlOptions [
  (''finiteModel'',xml1manyElements ''finiteModel''
    (xmlChange (xml2nat ''carrierSize'') (\<lambda> n :: nat. n - 1))
    (xml3elements ''interpret''
      xml2name
      (xml2nat ''arity'')
      xml2arith_fun
      (\<lambda> f a inter. ((f, a), inter)))
    (\<lambda>n xs. Some(SL_Inter n xs))),
  (''rootLabeling'',xml0elements ''rootLabeling'' None)
]) id"

datatype 'a poly = PVar nat | PNum 'a | PSum "'a poly list" | PMult "'a poly list"

fun sum_lpoly :: "('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> 'a list \<Rightarrow> 'a list \<Rightarrow> 'a list"
where "sum_lpoly _ [] ys = ys"
  | "sum_lpoly _ xs [] = xs"
  | "sum_lpoly pl (x # xs) (y # ys) = pl x y # sum_lpoly pl xs ys"


fun poly_to_lpoly :: "'a \<Rightarrow> 'a \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> 'a poly \<Rightarrow> (string,'a \<times> 'a list)emonad"
where "poly_to_lpoly _ _ _ _ (PNum i) = return (i,[])"
 | "poly_to_lpoly ze on pl ti (PVar x) = return (ze, replicate x ze @ [on])"
 | "poly_to_lpoly ze on pl ti (PSum [p]) = poly_to_lpoly ze on pl ti p"
 | "poly_to_lpoly ze on pl ti (PSum (p # q # ps)) = (
      poly_to_lpoly ze on pl ti p 
  \<guillemotright>= (\<lambda> (cp,ncp). poly_to_lpoly ze on pl ti (PSum (q # ps))
  \<guillemotright>= (\<lambda> (cq,ncq). return (pl cp cq,sum_lpoly pl ncp ncq))))"
 | "poly_to_lpoly ze on pl ti (PMult [p]) = poly_to_lpoly ze on pl ti p"
 | "poly_to_lpoly ze on pl ti (PMult (p # q # ps)) = (
      poly_to_lpoly ze on pl ti p
  \<guillemotright>= (\<lambda> (cp,ncp). poly_to_lpoly ze on pl ti (PMult (q # ps))
  \<guillemotright>= (\<lambda> (cq,ncq). if (forall (op = ze) ncp) then return (ti cp cq, map (ti cp) ncq)
             else (if (forall (op = ze) ncq) then return (ti cp cq, map (\<lambda> x. ti x cq) ncp)
             else error ''non-linear polynomials not supported''))))"

function xml2poly :: "(xml \<Rightarrow> (string,'a)emonad) \<Rightarrow> xml \<Rightarrow> (string,'a poly)emonad"
where "xml2poly xml2coeff x = xml1element ''polynomial'' 
   (xmlOptions [
      (''coefficient'',xml1element ''coefficient'' xml2coeff PNum), 
      (''variable'',xmlChange (xml2nat ''variable'') (\<lambda> n. PVar (n - 1))),
      (''sum'',xmlManyElements ''sum'' (xml2poly xml2coeff) PSum),
      (''product'',xmlManyElements ''product'' (xml2poly xml2coeff) PMult)
    ]) id x"
by pat_completeness auto

definition xml2int_coeff :: "xml \<Rightarrow> (string,int)emonad"
where "xml2int_coeff \<equiv> xml2int ''integer''"

definition xml2vec_coeff :: "(xml \<Rightarrow> (string,'a)emonad) \<Rightarrow> xml \<Rightarrow> (string,'a vec)emonad"
where "xml2vec_coeff xml2coeff \<equiv> xmlManyElements ''vector''
              (xml1element ''coefficient'' xml2coeff id) id"

(* allow vectors and convert them to matrices by filling in zeros *)
definition xml2mat_coeff :: "'a \<Rightarrow> (xml \<Rightarrow> (string,'a)emonad) \<Rightarrow> xml \<Rightarrow> (string,'a mat)emonad"
where "xml2mat_coeff ze xml2coeff \<equiv> xmlOptions [
  (''matrix'',xmlManyElements ''matrix'' (xml2vec_coeff xml2coeff) id),
  (''vector'',xmlChange (xml2vec_coeff xml2coeff) (\<lambda> v. v # replicate (length v - 1) (replicate (length v) ze)))
  ]"

definition xml2arctic_coeff :: "xml \<Rightarrow> (string,arctic)emonad"
where "xml2arctic_coeff \<equiv> xmlOptions [
  (''integer'',xmlChange (xml2int ''integer'') Num_arc),
  (''minusInfinity'',xml0elements ''minusInfinity'' MinInfty)
  ]"


datatype Domain = Natural | Arctic | Int_mat nat | Arc_mat nat | Rational rat | Rat_mat rat nat

definition default_delta :: rat where "default_delta = 1 / 1000000000000000000000000000000000000"


definition xml2interpretation_type :: "xml \<Rightarrow> (string,Domain)emonad" where
  "xml2interpretation_type \<equiv> 
      (xml1element ''type'' 
          (xmlOptions [
            (''polynomial'', (xml2elements ''polynomial'' 
               (xml1element ''domain'' 
                 (xmlOptions [
                     (''naturals'',xml0elements ''naturals'' Natural),
                     (''arctic'',xml0elements ''arctic'' Arctic),
                     (''arcticBelowZero'',xml0elements ''arcticBelowZero'' Arctic),
                     (''rationals'',xml0elements ''rationals'' (Rational default_delta)),
                     (''matrices'',xml3elements ''matrices''
                          (xml2nat ''dimension'') 
                          (xml2nat ''strictDimension'') 
                          (xml1element ''domain'' (xmlOptions [
                                  (''rationals'',xml0elements ''rationals'' (0 :: nat)),
                                  (''naturals'',xml0elements ''naturals'' (1 :: nat)),
                                  (''arctic'',xml0elements ''arctic'' 2),
                                  (''arcticBelowZero'',xml0elements ''arcticBelowZero'' 2)
                              ]) id)
                          (\<lambda> di sd do. if do = 1 then Int_mat di else if do = 0 then Rat_mat default_delta di else Arc_mat di))
                     ]) 
                 id)
             (xml2nat ''degree'') 
             (\<lambda> type _ . type))),
             (''matrixInterpretation'',xml3elements ''matrixInterpretation''
                          (xml1element ''domain'' (xmlOptions [
                                  (''rationals'',xml0elements ''rationals'' (0 :: nat)),
                                  (''naturals'',xml0elements ''naturals'' (1 :: nat)),
                                  (''arctic'',xml0elements ''arctic'' 2),
                                  (''arcticBelowZero'',xml0elements ''arcticBelowZero'' 2)
                              ]) id)
                          (xml2nat ''dimension'') 
                          (xml2nat ''strictDimension'') 
                          (\<lambda> do di sd. if do = 1 then Int_mat di else if do = 0 then Rat_mat default_delta di else Arc_mat di))                 
            ])
          id)"

datatype ('a,'b,'c,'d,'e,'f)alter = Alter1 'a | Alter2 'b | Alter3 'c | Alter4 'd | Alter5 'e | Alter6 'f 
fun alter1 where "alter1 (Alter1 x) = x"
fun alter2 where "alter2 (Alter2 x) = x"
fun alter3 where "alter3 (Alter3 x) = x"
fun alter4 where "alter4 (Alter4 x) = x"
fun alter5 where "alter5 (Alter5 x) = x"
fun alter6 where "alter6 (Alter6 x) = x"


definition xml2interpretation :: "(xml \<Rightarrow> (string,'a)emonad) \<Rightarrow> xml \<Rightarrow> (string, 'a redpair)emonad"
where "xml2interpretation xml2name \<equiv> 
      (xml1CmanyElements ''interpretation'' 
       xml2interpretation_type
       (\<lambda> type. case type of
          Natural \<Rightarrow> (\<lambda> c. (xml3elements ''interpret''
                        xml2name
                        (xml2nat ''arity'')
                        (\<lambda> x. xml2poly xml2int_coeff x \<guillemotright>= poly_to_lpoly (get_comp_0 c) (get_comp_1 c) (get_comp_plus c) (get_comp_times c))
                        (\<lambda> f a poly.  (Alter1 ((f,a),(fst poly, fitLength (get_comp_0 c) a (snd poly))))))) 
                    int_components
        | Rational d \<Rightarrow> (\<lambda> c. (xml3elements ''interpret''
                        xml2name
                        (xml2nat ''arity'')
                        (\<lambda> x. xml2poly xml2rat x \<guillemotright>= poly_to_lpoly (get_comp_0 c) (get_comp_1 c) (get_comp_plus c) (get_comp_times c))
                        (\<lambda> f a poly.  (Alter2 ((f,a),(fst poly, fitLength (get_comp_0 c) a (snd poly))))))) 
                    (rat_components d)
        | Arctic \<Rightarrow> (\<lambda> c. (xml3elements ''interpret''
                        xml2name
                        (xml2nat ''arity'')
                        (\<lambda> x. xml2poly xml2arctic_coeff x \<guillemotright>= poly_to_lpoly (get_comp_0 c) (get_comp_1 c) (get_comp_plus c) (get_comp_times c))
                        (\<lambda> f a poly.  (Alter3 ((f,a),(fst poly, fitLength (get_comp_0 c) a (snd poly))))))) 
                    arctic_components
        | Int_mat n \<Rightarrow> (\<lambda> c. (xml3elements ''interpret''
                        xml2name
                        (xml2nat ''arity'')
                        (\<lambda> x. xml2poly (xml2mat_coeff 0 xml2int_coeff) x \<guillemotright>= poly_to_lpoly (get_comp_0 c) (get_comp_1 c) (get_comp_plus c) (get_comp_times c))
                        (\<lambda> f a poly.  (Alter4 ((f,a),(fst poly, fitLength (get_comp_0 c) a (snd poly))))))) 
                    (int_mat_components (n - 1))
        | Rat_mat d n \<Rightarrow> (\<lambda> c. (xml3elements ''interpret''
                        xml2name
                        (xml2nat ''arity'')
                        (\<lambda> x. xml2poly (xml2mat_coeff 0 xml2rat) x \<guillemotright>= poly_to_lpoly (get_comp_0 c) (get_comp_1 c) (get_comp_plus c) (get_comp_times c))
                        (\<lambda> f a poly.  (Alter5 ((f,a),(fst poly, fitLength (get_comp_0 c) a (snd poly))))))) 
                    (rat_mat_components d (n - 1))
        | Arc_mat n \<Rightarrow> (\<lambda> c. (xml3elements ''interpret''
                        xml2name
                        (xml2nat ''arity'')
                        (\<lambda> x. xml2poly (xml2mat_coeff MinInfty xml2arctic_coeff) x \<guillemotright>= poly_to_lpoly (get_comp_0 c) (get_comp_1 c) (get_comp_plus c) (get_comp_times c))
                        (\<lambda> f a poly.  (Alter6 ((f,a),(fst poly, fitLength (get_comp_0 c) a (snd poly))))))) 
                    (arc_mat_components (n - 1))
        )
      (\<lambda> type pi. case type of 
            Natural \<Rightarrow> Int_carrier (map alter1 pi)
          | Rational d \<Rightarrow> Rat_carrier d (map alter2 pi)
          | Arctic \<Rightarrow> Arctic_carrier (map alter3 pi)
          | Int_mat n \<Rightarrow> Int_mat_carrier (n - 1) (map alter4 pi)
          | Rat_mat d n \<Rightarrow> Rat_mat_carrier d (n - 1) (map alter5 pi)
          | Arc_mat n \<Rightarrow> Arc_mat_carrier (n - 1) (map alter6 pi)
      ))
    "

definition xml2afs :: "(xml \<Rightarrow> (string,'a)emonad) \<Rightarrow> xml \<Rightarrow> (string, 'a afs_list)emonad"
where "xml2afs xml2name \<equiv> xmlManyElements ''argumentFilter'' 
  (xml3elements ''argumentFilterEntry''
      xml2name
      (xml2nat ''arity'')
      (xmlOptions [(''collapsing'',xmlChange (xml2nat ''collapsing'') (\<lambda> i. Collapse (i - 1))),
                   (''nonCollapsing'',xmlManyElements ''nonCollapsing'' xml2position AFList)])
      (\<lambda> f a e. ((f,a), e))
  ) id"

definition xml2precedence :: "(xml \<Rightarrow> (string,'a)emonad) \<Rightarrow> xml \<Rightarrow> (string, 'a lpo_repr)emonad"
where "xml2precedence xml2name \<equiv> xmlManyElements ''statusPrecedence'' 
  (xml4elements ''statusPrecedenceEntry''
      xml2name
      (xml2nat ''arity'')
      (xml2nat ''precedence'')
      (xml0elements ''lex'' ())
      (\<lambda> f a p s. ((f,a), p))
  ) id"

definition xml2path_order :: "(xml \<Rightarrow> (string,'a)emonad) \<Rightarrow> xml \<Rightarrow> (string, 'a redpair)emonad"
where "xml2path_order xml2name \<equiv> xml1to2elements ''pathOrder''
   (xml2precedence xml2name)
   (xml2afs xml2name)
   (\<lambda> pr afo. case afo of None \<Rightarrow> LPO pr [] | Some af \<Rightarrow> LPO pr af)"

definition xml2redpair :: "(xml \<Rightarrow> (string,'a)emonad) \<Rightarrow> xml \<Rightarrow> (string, 'a redpair)emonad"
where "xml2redpair xml2name = 
  xml1element ''redPair'' 
    (xmlOptions [(''interpretation'',xml2interpretation xml2name), (''pathOrder'',xml2path_order xml2name)]) id
   "

definition id_labeling :: "('f,'a,'a list)label" 
  where "id_labeling _ cs = Left cs" 

definition xml2scg :: "(xml \<Rightarrow> (string,'f)emonad) \<Rightarrow> xml \<Rightarrow> (string,('f,string)rule \<times> ((nat \<times> nat)list \<times> (nat \<times> nat)list))emonad"
where "xml2scg xml2name \<equiv> xml1manyElements ''sizeChangeGraph'' 
    (xml2rule xml2name)
    (xml3elements ''edge''
       xml2position
       (xml2bool ''strict'')
       xml2position
       (\<lambda> p s q. (s,(p,q)))
    )
    (\<lambda> lr edges. (lr,(map snd (filter fst edges), map snd (filter (\<lambda> x. \<not> fst x) edges))))"

function xml2dpProof :: "nat \<Rightarrow> (nat \<Rightarrow> xml \<Rightarrow> (string,('f,labelType)lab)emonad) \<Rightarrow> xml \<Rightarrow> (string,('f,labelType,string)DPProof)emonad"
where "xml2dpProof ld xml2name x = xmlOptions [ 
  (''pIsEmpty'',xml0elements ''pIsEmpty'' 
     PisEmpty),
  (''depGraphProc'',xmlManyElements ''depGraphProc''
     (xml2ndChoice ''component'' 
         (xml1element ''dps'' (xml2rules (xml2name ld)) id)
         ''realScc''
         (xml2dpProof ld xml2name)
         (\<lambda> dps prfOpt. (prfOpt, dps)))
     DepGraphProc),
  (''redPairProc'',xml3elements ''redPairProc'' 
     (xml2redpair (xml2name ld)) 
     (xml1element ''dps'' (xml2rules (xml2name ld)) id) 
     (xml2dpProof ld xml2name) 
     RedPairProc),
  (''redPairUrProc'',xml4elements ''redPairUrProc'' 
     (xml2redpair (xml2name ld)) 
     (xml1element ''dps'' (xml2rules (xml2name ld)) id) 
     (xml1element ''usableRules'' (xml2rules (xml2name ld)) id) 
     (xml2dpProof ld xml2name)
     RedPairUrProc),
  (''monoRedPairProc'',xml4elements ''monoRedPairProc'' 
     (xml2redpair (xml2name ld)) 
     (xml1element ''dps'' (xml2rules (xml2name ld)) id) 
     (xml1element ''trs'' (xml2rules (xml2name ld)) id) 
     (xml2dpProof ld xml2name)
     MonoRedPairProc),
  (''monoRedPairUrProc'',xml5elements ''monoRedPairUrProc'' 
     (xml2redpair (xml2name ld)) 
     (xml1element ''dps'' (xml2rules (xml2name ld)) id) 
     (xml1element ''trs'' (xml2rules (xml2name ld)) id) 
     (xml1element ''usableRules'' (xml2rules (xml2name ld)) id) 
     (xml2dpProof ld xml2name)
     MonoRedPairUrProc),
  (''semlabProc'',xml4elements ''semlabProc'' 
     (xml2sl_inter (xml2name ld)) 
     (xml1element ''dps'' (xml2rules (xml2name (Suc ld))) id) 
     (xml1element ''trs'' (xml2rules (xml2name (Suc ld))) id) 
     (xml2dpProof (Suc ld) xml2name)
     (\<lambda>i. case i of Some i \<Rightarrow> SemlabProc i id_labeling
                  | None   \<Rightarrow> RootlabProc)),
  (''unlabProc'',xml3elements ''unlabProc''
     (xml1element ''dps'' (xml2rules (xml2name (ld - 1))) id) 
     (xml1element ''trs'' (xml2rules (xml2name (ld - 1))) id) 
     (xml2dpProof (ld - 1) xml2name)
     UnlabProc),
  (''subtermProc'',xml3elements ''subtermProc''
     (xml2sp (xml2name ld))
     (xml1element ''dps'' (xml2rules (xml2name ld)) id)
     (xml2dpProof ld xml2name)
     SpscProc),
  (''sizeChangeCriterion'',xml1manyElements ''sizeChangeCriterion''
     (xmlOptions [
         (''subtermCriterion'', xml0elements ''subtermCriterion'' (None)),
         (''reductionPair'', xml2elements ''reductionPair'' 
              (xml2redpair (xml2name ld)) 
              (xml1element ''usableRules'' (xml2rules (xml2name ld)) id) 
              (\<lambda> redp ur. Some (redp,ur)))
     ])
     (xml2scg (xml2name ld))
     (\<lambda> version. case version of 
         None \<Rightarrow> SizeChangeSubTerm
       | Some redp_ur \<Rightarrow> SizeChangeRedPair (fst redp_ur) (snd redp_ur))
  )    
  ] x"
by (pat_completeness, auto)

definition xmlIgnore :: "string \<Rightarrow> xml \<Rightarrow> (string,unit)emonad"
where "xmlIgnore name xml \<equiv> if name = (xmlGetName xml) then return () else error (''expected '' @ name @ '' element, but found '' @ show xml)"

function xml2trsProof :: "(xml \<Rightarrow> (string,('f,labelType)lab)emonad) \<Rightarrow> (xml \<Rightarrow> (string,('f,labelType)lab shp)emonad) \<Rightarrow> (nat \<Rightarrow> xml \<Rightarrow> (string,(('f,labelType)lab shp,labelType)lab)emonad) \<Rightarrow> xml \<Rightarrow> (string,('f,labelType,string)TRSProof)emonad"
where "xml2trsProof xml2name xml2sname xml2lsname x = xmlOptions [
   (''rIsEmpty'',xml0elements ''rIsEmpty''
      RisEmpty),
   (''dpTrans'',xml3elements ''dpTrans'' 
      (xml1element ''dps'' (xml2rules xml2sname) id)
      (xml2bool ''markedSymbols'')
      (xml2dpProof 0 xml2lsname)
      (\<lambda> dps _ prf. DPTrans dps prf)),
   (''ruleRemoval'',xml3elements ''ruleRemoval''
      (xml2redpair xml2name) 
      (xml1element ''trs'' (xml2rules xml2name) id)
      (xml2trsProof xml2name xml2sname xml2lsname)
      RuleRemoval),
  (''unlab'',xml2elements ''unlab''
     (xml1element ''trs'' (xml2rules xml2name) id) 
     (xml2trsProof xml2name xml2sname xml2lsname)
     Unlab),
  (''semlab'',xml3elements ''semlab'' 
     (xml2sl_inter xml2name) 
     (xml1element ''trs'' (xml2rules xml2name) id) 
     (xml2trsProof xml2name xml2sname xml2lsname)
     (\<lambda>i. case i of Some i \<Rightarrow> Semlab i id_labeling
                  | None   \<Rightarrow> Rootlab)),
  (''flatContextClosure'',xml3elements ''flatContextClosure''
    (xml2flat_contexts xml2name)
    (xml1element ''trs'' (xml2rules xml2name) id)
    (xml2trsProof xml2name xml2sname xml2lsname)
    Fcc
  )
  ] x"
by (pat_completeness, auto)

definition xml2trsDisproof :: "(xml \<Rightarrow> (string,'f)emonad) \<Rightarrow> xml \<Rightarrow> (string,('f,string)TRSDisproof)emonad"
where "xml2trsDisproof xml2name = xmlOptions [
   (''variableCondition'',xml0elements ''variableCondition''
      NotWellFormed),
   (''loop'',xmlMany2Elements ''loop'' 
      (xml3elements ''redex'' (xml2term xml2name) xml2pos (xml2rule xml2name) (\<lambda> t p r. t))
      (xml2substitution xml2name)
      (xml2context xml2name)
      (\<lambda> ts \<sigma> C. Loop (ts, C, \<sigma>)))
   ]"

definition xml2trsInput :: "(xml \<Rightarrow> (string,'f)emonad) \<Rightarrow> xml \<Rightarrow> (string,('f,string)trsL)emonad"
where "xml2trsInput xml2name \<equiv> xml1element ''input''
   (xml1element ''trsInput''
      (xml1element ''trs'' (xml2rules xml2name) id)
      id)
   id"



definition terminationGoal :: xml
  where "terminationGoal \<equiv> XML ''goal'' [] [XML ''termination'' [] [] None] None"

definition nonTerminationGoal :: xml
  where "nonTerminationGoal \<equiv> XML ''goal'' [] [XML ''nonTermination'' [] [] None] None"

fun xml2certProblem :: "
  (xml \<Rightarrow> (string,'f)emonad) \<Rightarrow> 
  (xml \<Rightarrow> (string,('f,labelType)lab)emonad) \<Rightarrow> 
  (xml \<Rightarrow> (string,('f,labelType)lab shp)emonad) \<Rightarrow> 
  (nat \<Rightarrow> xml \<Rightarrow> (string,(('f,labelType)lab shp,labelType)lab)emonad) \<Rightarrow> 
  xml \<Rightarrow> (string,('f,labelType,string)certificationProblem)emonad"
where "xml2certProblem xml2uname xml2name xml2sname xml2lsname (XML name _ cs text) = (if name = ''certificationProblem'' \<and> text = None \<and> length cs \<ge> 3 then 
  xml2trsInput xml2uname (cs ! 1) \<guillemotright>= (\<lambda> trs. (if cs ! 0 = terminationGoal then 
           xml1element ''proof'' (xml2trsProof xml2name xml2sname xml2lsname) id (cs ! 2) \<guillemotright>= (\<lambda> p. return (CertificationProblem True trs (Left p)))
      else (if cs ! 0 = nonTerminationGoal then 
           xml1element ''proof'' (xml2trsDisproof xml2uname) id (cs ! 2) \<guillemotright>= (\<lambda> p. return (CertificationProblem False trs (Right p)))
      else error ''unknown goal type in certification problem''
   )))
  else error ''could not parse root element'')"

primrec xmldoc2certProblem :: "xmldoc \<Rightarrow> (string,(string,labelType,string)certificationProblem)emonad"
where "xmldoc2certProblem (XMLDOC header xml) = debug ''0'' ''parsing xml to certification problem'' (xml2certProblem (xml2text ''name'') xml2plainName xml2sharpName xml2labSharpName xml)"

fun parseCertProblem :: "(string,labelType,string)certificationProblem endParser"
where "parseCertProblem s = (parseXMLDoc s \<guillemotright>= xmldoc2certProblem)"

declare parseCertProblem.simps[simp del]

(* ... to xml *)

class xml =  
  fixes xml :: "'a \<Rightarrow> xml"

instantiation "term" :: ("shows","shows")xml
begin
fun xml_term 
  where "xml_term (Var x) = XML ''var'' [] [] (Some (show x))" 
      | "xml_term (Fun f ts) = XML ''funapp'' [] 
              (XML ''name'' [] [] (Some (show f)) # 
              (map (\<lambda> t. XML ''arg'' [] [xml_term t] None) ts)) 
           None"
instance ..
end


instantiation certificationProblem :: ("shows",type,"shows")xml
begin
fun xml_rules :: "('a,'c)trsL \<Rightarrow> xml"
where "xml_rules rls = XML ''rules'' [] (map (\<lambda> (l,r). XML ''rule'' [] [
         XML ''lhs'' [] [xml l] None,         
         XML ''rhs'' [] [xml r] None
         ] None) rls) None"
fun xml_certificationProblem :: "('a,'b,'c)certificationProblem \<Rightarrow> xml"
where "xml_certificationProblem (CertificationProblem term rls prf) = (let inp = 
  (XML ''input'' Nil 
     (XML ''trsInput'' Nil 
        (XML ''trs'' Nil (xml_rules rls # Nil) None # Nil) 
        None # Nil)
     None)
  in (let goal = if isLeft prf then terminationGoal else nonTerminationGoal 
  in XML ''problem'' Nil [goal,inp] None))"
instance ..
end




end
